/*
** ---------------------------------------------------------------
** t3d.c  -  Package to manage 3D matrix transformations.
**
** ---------------------------------------------------------------
** Reference:
**
** See class notes of courses CIV-2801 Fundamentos de Computacao
** Grafica Aplicada, Departamento de Engenharia Civil, PUC-Rio.
**
** ---------------------------------------------------------------
** Functionality:
**
** This package maintains a 3D transformation 4x4 matrix which can
** be initialized as the identity matrix, and which accumulates 
** the usual 3D geometric transformations and 3D visualization 
** transformations.
**
** There are two entry points to transform a given 3D point 
** according to the current transformation matrix: one dividing
** the coordinates of the transformed point by the fourth 
** homogeneous coordinate and the other returning the four 
** homogeneous coordinates.
**
** There is two entry points to transform a given 3D point according
** to the inverse of the current transformation matrix.
**
** There is one entry point to retrieve the current transformation
** matrix and another to retrieve its inverse.
**
** The 3D transformations considered are scaling, translation, 
** rotations, rotation of basis, camera transformation, 
** perspective projection, and view volume normalization.  
** There is an entry point for each one of these transformations.  
** The effect of these functions is to accumulate the corresponding 
** transformation in the current transformation matrix.
**
** ---------------------------------------------------------------
**
** Public functions:
** ---------------------------------------------------------------
**
** See header file "t3d.h".
**
** ---------------------------------------------------------------
**
** Created:    11-Nov-2006    Luiz Fernando Martha
**    Incomplete source code file.  Part of the objectives of 
**    this assignment is to complete the source code in
**    this file.
**
** Modified:   xx-Nov-2006    xxx
**    Report modifications here.
**      
*/

/*
** ---------------------------------------------------------------
** Global variables and symbols:
*/
#include <math.h>
#include "t3d.h"
#include "vec.h"

/*
** ---------------------------------------------------------------
** Local definitions and variables:
*/
#ifndef rad
#define rad(a)  ((a)*0.01745329252)
#endif

static Matrix3d mt =            /* Current transformation matrix */
 {
  { 1.0, 0.0, 0.0, 0.0 } ,
  { 0.0, 1.0, 0.0, 0.0 } ,
  { 0.0, 0.0, 1.0, 0.0 } ,
  { 0.0, 0.0, 0.0, 1.0 }
 };

/*
** ---------------------------------------------------------------
** Local functions:
*/

/*=======================  t3dAccumulate  ======================*/

/* This function pre-multiplies the given matrix to the current 
 * transformation matrix.
 */
static void t3dAccumulate( Matrix3d new_mt )
{
 double   t0, t1, t2;

 t0       =  mt[0][0]*new_mt[0][0] + mt[1][0]*new_mt[0][1] 
           + mt[2][0]*new_mt[0][2] + mt[3][0]*new_mt[0][3];
 t1       =  mt[0][0]*new_mt[1][0] + mt[1][0]*new_mt[1][1] 
           + mt[2][0]*new_mt[1][2] + mt[3][0]*new_mt[1][3];
 t2       =  mt[0][0]*new_mt[2][0] + mt[1][0]*new_mt[2][1] 
           + mt[2][0]*new_mt[2][2] + mt[3][0]*new_mt[2][3];
 mt[3][0] =  mt[0][0]*new_mt[3][0] + mt[1][0]*new_mt[3][1] 
           + mt[2][0]*new_mt[3][2] + mt[3][0]*new_mt[3][3];
 mt[2][0] =  t2;
 mt[1][0] =  t1;
 mt[0][0] =  t0;

 t0       =  mt[0][1]*new_mt[0][0] + mt[1][1]*new_mt[0][1] 
           + mt[2][1]*new_mt[0][2] + mt[3][1]*new_mt[0][3];
 t1       =  mt[0][1]*new_mt[1][0] + mt[1][1]*new_mt[1][1] 
           + mt[2][1]*new_mt[1][2] + mt[3][1]*new_mt[1][3];
 t2       =  mt[0][1]*new_mt[2][0] + mt[1][1]*new_mt[2][1] 
           + mt[2][1]*new_mt[2][2] + mt[3][1]*new_mt[2][3];
 mt[3][1] =  mt[0][1]*new_mt[3][0] + mt[1][1]*new_mt[3][1] 
           + mt[2][1]*new_mt[3][2] + mt[3][1]*new_mt[3][3];
 mt[2][1] =  t2;
 mt[1][1] =  t1;
 mt[0][1] =  t0;

 t0       =  mt[0][2]*new_mt[0][0] + mt[1][2]*new_mt[0][1] 
           + mt[2][2]*new_mt[0][2] + mt[3][2]*new_mt[0][3];
 t1       =  mt[0][2]*new_mt[1][0] + mt[1][2]*new_mt[1][1] 
           + mt[2][2]*new_mt[1][2] + mt[3][2]*new_mt[1][3];
 t2       =  mt[0][2]*new_mt[2][0] + mt[1][2]*new_mt[2][1] 
           + mt[2][2]*new_mt[2][2] + mt[3][2]*new_mt[2][3];
 mt[3][2] =  mt[0][2]*new_mt[3][0] + mt[1][2]*new_mt[3][1] 
           + mt[2][2]*new_mt[3][2] + mt[3][2]*new_mt[3][3];
 mt[2][2] =  t2;
 mt[1][2] =  t1;
 mt[0][2] =  t0;

 t0       =  mt[0][3]*new_mt[0][0] + mt[1][3]*new_mt[0][1] 
           + mt[2][3]*new_mt[0][2] + mt[3][3]*new_mt[0][3];
 t1       =  mt[0][3]*new_mt[1][0] + mt[1][3]*new_mt[1][1] 
           + mt[2][3]*new_mt[1][2] + mt[3][3]*new_mt[1][3];
 t2       =  mt[0][3]*new_mt[2][0] + mt[1][3]*new_mt[2][1] 
           + mt[2][3]*new_mt[2][2] + mt[3][3]*new_mt[2][3];
 mt[3][3] =  mt[0][3]*new_mt[3][0] + mt[1][3]*new_mt[3][1] 
           + mt[2][3]*new_mt[3][2] + mt[3][3]*new_mt[3][3];
 mt[2][3] =  t2;
 mt[1][3] =  t1;
 mt[0][3] =  t0;
}


/*
** ---------------------------------------------------------------
** Public functions:
*/

/*======================  t3dTransform  ========================*/

void t3dTransform( double *x, double *y, double *z )
{
 double r, s, t, h;

  r = mt[0][0]*(*x)+mt[0][1]*(*y)+mt[0][2]*(*z)+mt[0][3];
  s = mt[1][0]*(*x)+mt[1][1]*(*y)+mt[1][2]*(*z)+mt[1][3];
  t = mt[2][0]*(*x)+mt[2][1]*(*y)+mt[2][2]*(*z)+mt[2][3];
  h = mt[3][0]*(*x)+mt[3][1]*(*y)+mt[3][2]*(*z)+mt[3][3];
 *x = r/h;
 *y = s/h;
 *z = t/h;
}

/*=====================  t3dTransformW  ========================*/

void t3dTransformW( double *x, double *y, double *z, double *w )
{
 double r, s, t, h;

  r = mt[0][0]*(*x)+mt[0][1]*(*y)+mt[0][2]*(*z)+mt[0][3];
  s = mt[1][0]*(*x)+mt[1][1]*(*y)+mt[1][2]*(*z)+mt[1][3];
  t = mt[2][0]*(*x)+mt[2][1]*(*y)+mt[2][2]*(*z)+mt[2][3];
  h = mt[3][0]*(*x)+mt[3][1]*(*y)+mt[3][2]*(*z)+mt[3][3];
 *x = r;
 *y = s;
 *z = t;
 *w = h;
}

/*=====================  t3dTransformVec  ======================*/

void t3dTransformVec( double *x, double *y, double *z )
{
 double r, s, t;

  r = mt[0][0]*(*x)+mt[0][1]*(*y)+mt[0][2]*(*z);
  s = mt[1][0]*(*x)+mt[1][1]*(*y)+mt[1][2]*(*z);
  t = mt[2][0]*(*x)+mt[2][1]*(*y)+mt[2][2]*(*z);
 *x = r;
 *y = s;
 *z = t;
}

/* ====================  t3dGetTransform  ===================== */

void t3dGetTransform( Matrix3d mtx )
{
 int i, j;

 for( i=0; i<4; i++ )
   for( j=0; j<4; j++ )
     mtx[i][j] = mt[i][j];
}

/*======================  t3dIdentity  ========================= */

void t3dIdentity( void )
{
 mt[0][0] = 1.0; mt[0][1] = 0.0; mt[0][2] = 0.0; mt[0][3] = 0.0;
 mt[1][0] = 0.0; mt[1][1] = 1.0; mt[1][2] = 0.0; mt[1][3] = 0.0;
 mt[2][0] = 0.0; mt[2][1] = 0.0; mt[2][2] = 1.0; mt[2][3] = 0.0;
 mt[3][0] = 0.0; mt[3][1] = 0.0; mt[3][2] = 0.0; mt[3][3] = 1.0;
}

/*=========================  t3dScale  =========================*/

void t3dScale( double sx, double sy, double sz )
{
 Matrix3d scale;

/*** COMPLETE HERE: 01 ***/

 t3dAccumulate( scale );
}

/*=======================  t3dTranslate  =======================*/

void t3dTranslate( double tx, double ty, double tz )
{
 Matrix3d transl;

/*** COMPLETE HERE: 02 ***/

 t3dAccumulate( transl );
}

/*=========================  t3dRotate  ========================*/

void t3dRotate( double angle, char axis )
{
 double cost = cos(rad(angle));
 double sint = sin(rad(angle));
 Matrix3d rot;

 switch( axis )
 {
  case 'x':
   rot[0][0] = 0.0;    rot[0][1] = 0.0;    rot[0][2] = 1.0;    rot[0][3] = 0.0;
   rot[1][0] = 0.0;    rot[1][1] =  cost;  rot[1][2] = -sint;  rot[1][3] = 0.0;
   rot[2][0] = 0.0;    rot[2][1] =  sint;  rot[2][2] =  cost;  rot[2][3] = 0.0;
   rot[3][0] = 0.0;    rot[3][1] = 0.0;    rot[3][2] = 0.0;    rot[3][3] = 1.0;
   break;
  case 'y':
/*** COMPLETE HERE: 03 ***/
   break;
  default:
  case 'z':
   rot[0][0] =  cost;  rot[0][1] = -sint;  rot[0][2] = 0.0;    rot[0][3] = 0.0;
   rot[1][0] =  sint;  rot[1][1] =  cost;  rot[1][2] = 0.0;    rot[1][3] = 0.0;
   rot[2][0] = 0.0;    rot[2][1] = 0.0;    rot[2][2] = 1.0;    rot[2][3] = 0.0;
   rot[3][0] = 0.0;    rot[3][1] = 0.0;    rot[3][2] = 0.0;    rot[3][3] = 1.0;
   break;
 }

 t3dAccumulate( rot );
}

/*======================  t3dRotateVecOrg  =====================*/

void t3dRotateVecOrg( double angle, double vx, double vy, double vz )
{
 double   cost = cos(rad(angle));
 double   sint = sin(rad(angle));
 Matrix3d rot;
 Vector   v;                       /* normalized vector         */

 v.x = vx;
 v.y = vy;
 v.z = vz;
 vecUnit( &v, &v );

 rot[0][0] = v.x*v.x + cost*(1.0 - v.x*v.x);
 rot[0][1] = v.x*v.y*(1.0 - cost) - v.z*sint;
 rot[0][2] = v.z*v.x*(1.0 - cost) + v.y*sint;
 rot[0][3] = 0.0;
 rot[1][0] = v.x*v.y*(1.0 - cost) + v.z*sint;
 rot[1][1] = v.y*v.y + cost*(1.0 - v.y*v.y);
 rot[1][2] = v.y*v.z*(1.0 - cost) - v.x*sint;
 rot[1][3] = 0.0;
 rot[2][0] = v.z*v.x*(1.0 - cost) - v.y*sint;
 rot[2][1] = v.y*v.z*(1.0 - cost) + v.x*sint;
 rot[2][2] = v.z*v.z + cost*(1.0 - v.z*v.z);
 rot[2][3] = 0.0;
 rot[3][0] = 0.0;
 rot[3][1] = 0.0;
 rot[3][2] = 0.0;
 rot[3][3] = 1.0;

 t3dAccumulate( rot );
}

/*======================  t3dRotateVecPnt  =====================*/

void t3dRotateVecPnt( double angle, double vx, double vy, double vz,
                      double px, double py, double pz )
{
/* Translate to origin.
 */
 t3dTranslate( -px, -py, -pz );

/* Rotate about given vector direction.
 */
 t3dRotateVecOrg( angle, vx, vy, vz );

/* Translate back to original position.
 */
 t3dTranslate( px, py, pz );
}

/*======================  t3dRotateBasis  ======================*/

void t3dRotateBasis( double ux, double uy, double uz, 
                     double vx, double vy, double vz,
                     double wx, double wy, double wz )
{
 Matrix3d rot;

/*** COMPLETE HERE: 04 ***/

 t3dAccumulate( rot );
}

/*=========================  t3dCamera  ========================*/

void t3dCamera( double eyex, double eyey, double eyez, 
                double refx, double refy, double refz,
                double vupx, double vupy, double vupz )
{
 Vector  view, upvec;   /* view vector and view upper vector */
 Vector  u, v, w;       /* eye system unit vectors           */

/* Define view vector (from eye to ref.) and view up vector.
 */
/*** COMPLETE HERE: 05 ***/

/* Find the eye system unit vectors.
 */
/*** COMPLETE HERE: 06 ***/

/* Concatenate a translation of basis from origin to eye point.
 */
/*** COMPLETE HERE: 07 ***/

/* Concatenate a rotation of basis from world to eye system.
 */
/*** COMPLETE HERE: 08 ***/
}

/*======================  t3dPerspective  ======================*/

void t3dPerspective( double d, double front, double back )
{
 double   alpha, beta;        /* perspective parameters */
 Matrix3d persp;

/*** COMPLETE HERE: 09 ***/

 t3dAccumulate( persp );
}

/*=====================  t3dNormalizeView  =====================*/

void t3dNormalizeView( double left,   double right, 
                       double bottom, double top, 
                       double front,  double back )
{
 double tx, ty, tz;           /* translation factors */
 double sx, sy, sz;           /* scaling factors     */

/* Concatenate a translation of basis from origin to center
 * of view volume.
 */
/*** COMPLETE HERE: 10 ***/
 t3dTranslate( tx, ty, tz );

/* Concatenate a scaling transformation that scales the view 
 * volume to a cube ranging from -1 to +1 in the 3 directions, 
 * also reverting the z-axis direction.
 */
/*** COMPLETE HERE: 11 ***/
 t3dScale( sx, sy, sz );
}
