/*
** ---------------------------------------------------------------
** t3d.h: Definitions and prototypes for package to manage 3D
**        matrix transformations.
**
** ---------------------------------------------------------------
** Reference:
**
** See class notes of courses CIV-2801 Fundamentos de Computacao
** Grafica Aplicada, Departamento de Engenharia Civil, PUC-Rio.
**
** ---------------------------------------------------------------
** Functionality:
**
** This package maintains a 3D transformation 4x4 matrix which can
** be initialized as the identity matrix, and which accumulates 
** the usual 3D geometric transformations and 3D visualization 
** transformations.
**
** There are two entry points to transform a given 3D point 
** according to the current transformation matrix: one dividing
** the coordinates of the transformed point by the fourth 
** homogeneous coordinate and the other returning the four 
** homogeneous coordinates.
**
** There is one entry point to retrieve the current transformation
** matrix and another to retrieve its inverse.
**
** The 3D transformations considered are scaling, translation, 
** rotations, rotation of basis, camera transformation, 
** perspective projection, and view volume normalization.  
** There is an entry point for each one of these transformations.  
** The effect of these functions is to accumulate the corresponding 
** transformation in the current transformation matrix.
**
** ---------------------------------------------------------------
**
** Public functions:
** ---------------------------------------------------------------
**
** void t3dTransform( double* x, double* y, double* z )
**
**    x, y, z    - coordinates of transformed point       (in/out)
**
** This function transforms the given coordinates according
** to the current transformation matrix.
** The result of the transformation is divided by the fourth
** homogeneous coordinate.
**
** ---------------------------------------------------------------
**
** void t3dTransformW( double* x, double* y, double* z, double *w )
**
**    x, y, z    - homogeneous coords. of transformed pt. (in/out)
**    w          - fourth homogeneous coord. of transf. pt.  (out)
**
** This function transforms the given coordinates according
** to the current transformation matrix.
** The four homogeneous coordinates of the transformed point are
** returned, that is, the result of the transformation is NOT 
** divided by the fourth homogeneous coordinate.
**
** ---------------------------------------------------------------
**
** void t3dTransformVec( double* x, double* y, double* z )
**
**    x, y, z  - direction coords. of transformed vector. (in/out)
**
** This function transforms the given vector according to the 
** current transformation matrix.
** The three direction coordinates of the transformed vector are
** returned. It is assumed that the result of the transformation is 
** also a direction, that is the four homogeneous coordinate is 
** zero.
**
** ---------------------------------------------------------------
**
** void t3dGetTransform( Matrix3d mtx )
**
**    mtx    - current transformation matrix                 (out)
**
** This function returns the current transformation matrix.
**
** ---------------------------------------------------------------
**
** void t3dIdentity( void )
**
** This function sets the current transformation matrix and its 
** inverse as identity matrices.
**
** ---------------------------------------------------------------
**
** void t3dScale( double sx, double sy, double sz )
**
**    sx, sy, sz - scaling parameters                         (in)
**
** This function accumulates a scaling transformation to the
** current transformation matrix.
**
** ---------------------------------------------------------------
**
** void t3dTranslate( double tx, double ty, double tz )
**
**    tx, ty, tz  - translation parameters                    (in)
**
** This function accumulates a translation transformation to
** the current transformation matrix.
**
** ---------------------------------------------------------------
**
** void t3dRotate( double angle, char axis )
**
**    angle    - rotation angle (degrees)                     (in)
**    axis     - axis of rotation (either 'x', 'y', or 'z')   (in)
**
** This function accumulates a rotation transformation to the
** current transformation matrix.  The rotation is performed 
** about the origin and around one of the cartesian axis, which is 
** indicated by the second parameter.  The rotation angle in 
** assumed given in degrees.
**
** ---------------------------------------------------------------
**
** void t3dRotateVecOrg( double angle, double vx, double vy, double vz )
**
**    angle    - rotation angle (degrees)                     (in)
**    vx,vy,vz - vector direction coordinates                 (in)
**
** This function accumulates a rotation transformation to the
** current transformation matrix.  The rotation is performed
** around a given vector through the origin.  The vector direction
** is indicated by its coordinates.  The rotation angle in assumed
** given in degrees.
**
** ---------------------------------------------------------------
**
** void t3dRotateVecPnt( double angle, double vx, double vy, double vz,
**                       double px, double py, double pz )
**
**    angle    - rotation angle (degrees)                     (in)
**    vx,vy,vz - vector direction coordinates                 (in)
**    px,py,pz - point coordinates                            (in)
**
** This function accumulates a rotation transformation to the
** current transformation matrix.  The rotation is performed
** around a given vector through a given point.  The vector
** direction and the point of reference for rotation are indicated
** by their coordinates.  The rotation angle in assumed given in
** degrees.
**
** ---------------------------------------------------------------
**
** void t3dRotateBasis( double ux, double uy, double uz, 
**                      double vx, double vy, double vz,
**                      double wx, double wy, double wz )
**
**    ux,uy,uz,
**    vx,vy,vz,
**    wx,wy,wz - coordinates of unitary vector directions of 
**               the three new basis axes.                    (in)
**
** This function accumulates a rotation of basis transformation
** to the current transformation matrix.
** The rotation of basis is a rotation transformation whose matrix
** is such that each row of the 3x3 linear transformation submatrix
** is formed by the given new basis unitary vector coordinates.
**
** ---------------------------------------------------------------
**
** void t3dCamera( double eyex, double eyey, double eyez, 
**                 double refx, double refy, double refz,
**                 double vupx, double vupy, double vupz )
**
**    eyex, eyey, eyez     - eye position                     (in)
**    refx, refy, refz     - reference point position         (in)
**    vupx, vupy, vupz     - view up vector direction         (in)
**
** This function accumulates a camara transformation to the
** current transformation matrix.  The camara transformation is
** a change-of-basis transformation formed by a concatenation of
** a translation from the origin of the world (client) coordinate
** sytem to the eye point and a rotation from the world coordinate 
** system to the eye coordinate system (a right-handed system 
** centered at the eye point, with the z axis from the reference
** point to the eye point).
** The given view upper vector defines with the view direction 
** (eye system z axis, going from the reference point to the eye
** point) the eye system yz plane.
**
** ---------------------------------------------------------------
**
** void t3dPerspective( double d, double front, double back )
**
**    d           -  distance of projection plane to origin   (in)
**    front       -  frustum front clipping plane position    (in)
**    back        -  frustum back clipping plane position     (in)
**
** This function accumulates a canonical perspective 
** transformation to the current transformation matrix.  This
** transformation defines a conical projection whose center of
** projection is assumed as the origin of the system of 
** coordinates.  Alternatively, this transformation defines a 
** distortion of the coordinate space such that the result of a 
** orthographic projection of the point in the distorted space 
** coincides with the result of the conical projection.
** The plane of projection and the two frustum clipping planes 
** (front and back) are assumed normal to the z axis, lying on 
** its negative side.  These planes are defined by their 
** distances to the origin.  
** It is assumed that both values for the clipping plane
** positions are positive and that the visible part of the
** 3D space lies along the negative side of the z-axis in 
** the eye coordinate system.
** The perspective distortion is such that it does not change 
** z coordinate values of points lying on the two frustum 
** clipping planes.
**
** ---------------------------------------------------------------
**
** void t3dNormalizeView( double left, double right, 
**                        double bottom, double top, 
**                        double front, double back )
**
**    left, right - horizontal clipping limits of view volume (in)
**    bottom, top - vertical clipping limits of view volume   (in)
**    front, back - deepness clipping limits of view volume   (in)
**
** This function accumulates a view volume normalization 
** transformation to the current transformation matrix.  This
** normalization is such that it transforms a view volume defined
** in the eye (camera) coordinate system into a cube whose center
** is at the origin and whose coordinates are in the range from
** -1 to +1.
** The given view volume defines, in the eye coordinate system, 
** the portion of modeling space which is seen on the screen.  
** That is, all points lying ouside the view volume are clipped.
** It is assumed that both values for the deepness clipping 
** plane positions are positive and that the visible part of the
** 3D space lies along the negative side of the z-axis in 
** the eye coordinate system.  The normalization transformation 
** is such that points at the front clipping plane will have z-
** coordinates equal to -1 and points at the back clipping plane 
** will have z-coordinates equal to +1.  Therefore, there is a
** reflection transformation with respect to the xy plane in the 
** view volume normalization transformation.
**
** ---------------------------------------------------------------
**
** Created:    08-Nov-2006    Luiz Fernando Martha
**      Stolen from G3d library.
**
*/

#ifndef _T3D_H
#define _T3D_H 

/* Matrix type
 */
typedef double Matrix3d[4][4];

/* Prototypes of functions
 */

void t3dTransform( double *x, double *y, double *z );
void t3dTransformW( double *x, double *y, double *z, double *w );
void T3dTransformVec( double *x, double *y, double *z );
void t3dGetTransform( Matrix3d mtx );
void t3dIdentity( void );
void t3dScale( double sx, double sy, double sz );
void t3dTranslate( double tx, double ty, double tz );
void t3dRotate( double angle, char axis );
void t3dRotateVecOrg( double angle, double vx, double vy, double vz );
void t3dRotateVecPnt( double angle, double vx, double vy, double vz,
                      double px, double py, double pz );
void t3dRotateBasis( double ux, double uy, double uz, 
                     double vx, double vy, double vz,
                     double wx, double wy, double wz );
void t3dCamera( double eyex, double eyey, double eyez, 
                double refx, double refy, double refz,
                double vupx, double vupy, double vupz );
void t3dPerspective( double d, double front, double back );
void t3dNormalizeView( double left, double right,
                       double bottom, double top,
                       double front, double back );

#endif
