/*
** ---------------------------------------------------------------
** v3d.c - 3D View Specification Package
**
** ---------------------------------------------------------------
** Purpose:
** Allows incremental updating of the view of a model on a
** canvas window.
**
** See description and public function API in header file v3d.h.
**
** ---------------------------------------------------------------
**
** Version:    1.0
**
** Created:    11-Nov-2006    Luiz F. Martha
**    Stolen from V3d module (Tecgraf/PUC-Rio).
**    Incomplete source code file.  Part of the objectives of 
**    this assignment is to complete the source code in
**    this file.
**
** Modified:   xx-Nov-2006    xxx
**    Report modifications here.
**
*/

#define _V3D_C

/*
** ---------------------------------------------------------------
** Global variables and symbols
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "iup.h"
#include "iupkey.h"
#include "v3d.h"
#include "t3d.h"
#include "vec.h" 

/*
** ---------------------------------------------------------------
** Private variables and symbols:
*/
 
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#ifndef NULL
#define NULL 0L
#endif
#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif
#ifndef ABS
#define ABS(a) (((a)<0)?-(a):(a))
#endif

#ifndef PI
#define PI           3.141592653589793
#endif
#ifndef rad
#define rad(a)       ((a)*(PI/180.0))
#endif
#ifndef deg
#define deg(a)       ((a)*(180.0/PI))
#endif
#define Sqrt3        (1.7320508)
#define Sqrt6        (2.4494897)
#ifndef Len2
#define Len2(x,y)    sqrt(((x)*(x))+((y)*(y)))
#endif
#ifndef Len3
#define Len3(x,y,z)  sqrt(((x)*(x))+((y)*(y))+((z)*(z)))
#endif

#define DefaultIncr  0.02

#define ScaleFactor  1.0
#define RotFactor    1.0

#define FitFactor    1.15
#define ZoomFactor   1.20

/* Macro definitions for mouse configuration:
 * r -> 0='S'(shift) 1='C'(control) 2='1'(bt1) 3='1'(bt2) 4='1'(bt3)
 */
#define ShiftPressed(r)         (r[0]=='S')
#define ControlPressed(r)       (r[1]=='C')

#define ButtonPressed         0
#define ButtonReleased        1

/* Types of view specification mouse-controlled updates:
 */
typedef enum _v3dupdatetype {

/* Scaling of view window update */
  V3dScaleWindow,

/* Pan (2D) window update */
  V3dPanWindow,

/* Perspective distortion (eye distance) update */
  V3dEyeDistance,

/* Screen-based updates */
  V3dScrRotate,       /* Rotation about screen axes    */
  V3dScrSpin,         /* Spin (twist) rotation         */
  V3dScrTranslateUV,  /* Translation along screen axes */
  V3dScrTranslateW,   /* Axial translation (fwrd/back) */

/* Walk-through updates */
  V3dWlkRotate,       /* Look up, down, left, or right */
  V3dWlkSpin,         /* Spin camera around its axis   */
  V3dWlkTranslateUV,  /* Move up, down, left, or right */
  V3dWlkTranslateW,   /* Move forward and backward     */

  V3dNoUpdate

 } V3dUpdateType;

typedef struct _modelbox {  /* model bounding box descriptor	   */
	double	xmn, xmx;   /* lateral box limits */
	double	ymn, ymx;   /* height box limits */
	double	zmn, zmx;   /* deep box limits */
	} ModelBox;

/* Private data view spec. data descriptor:
 */
typedef struct _v3dstate {
  struct _v3dstate *next;
  void      *canvas;           /* client handle to target canvas*/
  int       hsize;             /* canvas width in pixels       */
  int       vsize;             /* canvas height in pixels      */
  Point     eye;               /* eye (camera) position        */
  Point     ref;               /* reference position           */
  Vector    vup;               /* up-vector (is a unit vector) */
  double    left;              /* left view volume limit       */
  double    right;             /* right view volume limit      */
  double    bottom;            /* bottom view volume limit     */
  double    top;               /* top view volume limit        */
  double    front;             /* front/back clip planes:      */
  double    back;              /* - measured as dist. from eye */
  char      defvup;            /* default view up-vector:      */
                               /* - either 'x', 'y', or 'z'    */
  char      type;              /* 'o' ==> orthographic         */
                               /* 'p' ==> perspective          */
  char      model;             /* 's' ==> screen based         */
                               /* 'w' ==> walk through model   */
                               /* 'o' ==> object based         */
  char      control;           /* 'r' ==> rotation control     */
                               /* 't' ==> translation control  */
  ModelBox  box;               /* client model bounding box    */

  /* Saved viewing parameters for restoring later: */
  int       saved;
  char      type_sav;
  Point     eye_sav;
  Point     ref_sav;
  Vector    vup_sav;
  double    left_sav;
  double    right_sav;
  double    bottom_sav;
  double    top_sav;
  double    front_sav;
  double    back_sav;

  /* Holds current view spec. update type: */
  V3dUpdateType  update_type;

  /* Camera transformation parameters: */
  Vector   rot_axis;          /* axis of rotation    */
  double   angle;             /* angle of rotation   */
  Point    center;            /* center of rotation  */
  Vector   transl;            /* translation vector  */

 } V3dState;


/* Package private specification data structure:
 */
typedef struct _v3dspec {

  /* Head of view specification descriptor list: */
  V3dState  *view_head;

  /* flag to change view (it is activated when mouse botton is pressed): */
  int change_view;

  /* Holds setup mouse position: */
  int setuppt_x, setuppt_y;

  /* Holds previous mouse position: */
  int prevpt_x, prevpt_y;

  /* Eye coordinate system: */
  Vector   u, v, w;     /* local system at eye point      */

 } V3dSpec;

/* Private data structure handle:
 */
static V3dSpec  v3d;

/*
** ---------------------------------------------------------------
** Prototype of private functions:
*/
static V3dState *GetViewState( void *canvas );
static void AdjustWindow( V3dState *view, double *lt, double *rt, 
                                          double *bt, double *tp );
static void BuildEyeSystem( V3dState *view );
static void ViewRotate( V3dState *view );
static void ViewTranslate( V3dState *view );
static void ViewScrRotU( V3dState *view, double incr );
static void ViewScrRotV( V3dState *view, double incr );
static void ViewScrRotW( V3dState *view, double incr );
static void ViewScrTranU( V3dState *view, double incr );
static void ViewScrTranV( V3dState *view, double incr );
static void ViewScrTranW( V3dState *view, double incr );
static void ViewWlkRotU( V3dState *view, double incr );
static void ViewWlkRotV( V3dState *view, double incr );
static void ViewWlkRotW( V3dState *view, double incr );
static void ViewWlkTranU( V3dState *view, double incr );
static void ViewWlkTranV( V3dState *view, double incr );
static void ViewWlkTranW( V3dState *view, double incr );
static void ViewObjRotX( V3dState *view, double incr );
static void ViewObjRotY( V3dState *view, double incr );
static void ViewObjRotZ( V3dState *view, double incr );
static void ViewObjTranX( V3dState *view, double incr );
static void ViewObjTranY( V3dState *view, double incr );
static void ViewObjTranZ( V3dState *view, double incr );
static void SetupMouseControl( V3dState *view, int b, int crtl, int shift );
static void ViewUpdateMouseMv( V3dState *view, double dx, double dy );
static void ViewScaleMouse( V3dState *view, double dx, double dy );
static void ViewPanMouse( V3dState *view, double dx, double dy );
static void ViewEyeDistMouse( V3dState *view, double dx, double dy );
static void ViewScrRotMouse( V3dState *view, double dx, double dy );
static void ViewScrSpinMouse( V3dState *view, double dx, double dy );
static void ViewScrTranUVMouse( V3dState *view, double dx, double dy );
static void ViewScrTranWMouse( V3dState *view, double dx, double dy );
static void ViewWlkRotMouse( V3dState *view, double dx, double dy );
static void ViewWlkSpinMouse( V3dState *view, double dx, double dy );
static void ViewWlkTranUVMouse( V3dState *view, double dx, double dy );
static void ViewWlkTranWMouse( V3dState *view, double dx, double dy );
static void ProjectBox( V3dState *view, double *left, double *right,
                                        double *bottom, double *top );

/*
** ---------------------------------------------------------------
** Private package functions:
*/

/* =======================  GetViewState  ====================== */

static V3dState *GetViewState( void *canvas )
{
 V3dState   *view;

 for( view = v3d.view_head; view != NULL; view = view->next )
 {
  if( view->canvas == canvas ) break;
 }
 return( view );
}

/* ========================  AdjustWindow  ======================= */

static void AdjustWindow( V3dState *view, double *lt, double *rt, 
                                          double *bt, double *tp )
{
 double  vpr;               /* viewport distortion ratio        */
 double  xc, yc;            /* center of window in world coords.*/
 double  hsize, vsize;      /* window sizes in world coords.    */

 vpr = (double)view->vsize / (double)view->hsize;

/* Find window center and sizes.
 */
 xc = (*rt + *lt) * 0.5;
 yc = (*tp + *bt) * 0.5;
 hsize = (*rt - *lt);
 vsize = (*tp - *bt);
 
/* Define window dimensions based on given lateral dimensions 
 * of the visualization volume.  This window is adjusted to
 * keep the same aspect ratio of the viewport.
 */
 if( vsize > (vpr*hsize) )
  hsize = vsize / vpr;
 else
  vsize = hsize * vpr;

 hsize *= 0.5;
 vsize *= 0.5;

 *lt = xc - hsize;
 *rt = xc + hsize;
 *bt = yc - vsize;
 *tp = yc + vsize;
}

/* ======================  BuildEyeSystem  ===================== */
/**
 * Compute unit vectors of eye coordinate system:
 * eye x = U = v3d.u
 * eye y = V = v3d.v
 * eye z = W = v3d.w
 */
static void BuildEyeSystem( V3dState *view )
{
 Vector view_vec;

/* Define view vector (from eye to ref.).
 */
 view_vec.x = view->ref.x - view->eye.x;
 view_vec.y = view->ref.y - view->eye.y;
 view_vec.z = view->ref.z - view->eye.z;

/* Find the eye system unit vectors.
 */
 vecUnit( &view_vec, &v3d.w );
 v3d.w.x *= -1.0;
 v3d.w.y *= -1.0;
 v3d.w.z *= -1.0;
 vecCross( &view->vup, &v3d.w, &v3d.u );
 vecUnit( &v3d.u, &v3d.u );
 vecCross( &v3d.w, &v3d.u, &v3d.v );
}

/* =========================  ViewRotate  ====================== */

static void ViewRotate( V3dState *view )
{
/* Rotate eye position and reference pts. about current center of
 * rotation around current rotation axis.
 */
 t3dIdentity( );
/*** COMPLETE HERE: 12 ***/

/* Rotate view up-vector around current rotation axis.
 */
 t3dIdentity( );
/*** COMPLETE HERE: 13 ***/
}

/* =======================  ViewTranslate  ===================== */

static void ViewTranslate( V3dState *view )
{
/* Translate reference and eye points in current translation direction.
 */
/*** COMPLETE HERE: 14 ***/
}

/* ========================  ViewScrRotU  ====================== */
/**
 * Turn object down and up around screen horizontal axis.
 */
static void ViewScrRotU( V3dState *view, double incr )
{
/* Set the axis of rotation: eye U axis.
 */
 view->rot_axis = v3d.u;

/* Set center of rotation at reference point.
 */
 view->center = view->ref;

/* Set the rotation angle based on given rotation increment.
 * Screen-based ==> rotate eye contrary to target object rotation.
 */
 view->angle = -incr;

/* Rotate eye position about reference point around found 
 * rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* ========================  ViewScrRotV  ====================== */
/**
  * Turn object left and right around screen vertical axis.
  */
static void ViewScrRotV( V3dState *view, double incr )
{
/* Set the axis of rotation: eye V axis.
 */
 view->rot_axis = v3d.v;

/* Set center of rotation at reference point.
 */
 view->center = view->ref;

/* Set the rotation angle based on given rotation increment.
 * Screen-based ==> rotate eye contrary to target object rotation.
 */
 view->angle = -incr;

/* Rotate eye position about reference point around found 
 * rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* ========================  ViewScrRotW  ====================== */
/**
  * Spin object around screen out-of-plane axis.
  */
static void ViewScrRotW( V3dState *view, double incr )
{
/* Set the axis of rotation: eye W axis.
 */
 view->rot_axis = v3d.w;

/* Set center of rotation at reference point.
 */
 view->center = view->ref;

/* Set the rotation angle based on given rotation increment.
 * Screen-based ==> rotate eye contrary to target object rotation.
 */
 view->angle = -incr;

/* Rotate eye position about reference point around found 
 * rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* ========================  ViewScrTranU  ===================== */
/**
 * Move object left and right along screen horizontal axis.
 */
static void ViewScrTranU( V3dState *view, double incr )
{
 double     min_size;          /* minimum window size            */
 double     tran_incr;         /* translation increment          */

/* Obtain translation increment in eye system at projection plane.
 * Screen-based ==> move eye and ref. contrary to target object translation.
 */
 min_size = MIN( view->right - view->left, view->top - view->bottom );
 tran_incr = -(min_size * incr);

/* Translate reference and eye points in eye U direction.
 */
 view->transl.x = tran_incr * v3d.u.x;
 view->transl.y = tran_incr * v3d.u.y;
 view->transl.z = tran_incr * v3d.u.z;
 ViewTranslate( view );
}

/* ========================  ViewScrTranV  ===================== */
/**
 * Move object down and up along screen vertical axis.
 */
static void ViewScrTranV( V3dState *view, double incr )
{
 double     min_size;          /* minimum window size            */
 double     tran_incr;         /* translation increment          */

/* Obtain translation increment in eye system at projection plane.
 * Screen-based ==> move eye and ref. contrary to target object translation.
 */
 min_size = MIN( view->right - view->left, view->top - view->bottom );
 tran_incr = -(min_size * incr);

/* Translate reference and eye points in eye V direction.
 */
 view->transl.x = tran_incr * v3d.v.x;
 view->transl.y = tran_incr * v3d.v.y;
 view->transl.z = tran_incr * v3d.v.z;
 ViewTranslate( view );
}

/* ========================  ViewScrTranW  ===================== */
/**
 * Move object forward and back along screen out-of-plane axis.
 */
static void ViewScrTranW( V3dState *view, double incr )
{
 double     min_size;          /* minimum window size            */
 double     tran_incr;         /* translation increment          */

/* Obtain translation increment in eye system at projection plane.
 * Screen-based ==> move eye and ref. contrary to target object translation.
 * Foward movement is in -W direction (sign changes twice).
 */
 min_size = MIN( view->right - view->left, view->top - view->bottom );
 tran_incr = min_size * incr;

/* Translate reference and eye points in eye W direction.
 */
 view->transl.x = tran_incr * v3d.w.x;
 view->transl.y = tran_incr * v3d.w.y;
 view->transl.z = tran_incr * v3d.w.z;
 ViewTranslate( view );
}

/* ========================  ViewWlkRotU  ====================== */
/**
 * Turn head down and up around screen horizontal axis.
 */
static void ViewWlkRotU( V3dState *view, double incr )
{
/* Set the axis of rotation: eye U axis.
 */
 view->rot_axis = v3d.u;

/* Set center of rotation at eye point.
 */
 view->center = view->eye;

/* Set the rotation angle based on given rotation increment.
 */
 view->angle = incr;

/* Rotate reference point about eye around found 
 * rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* ========================  ViewWlkRotV  ====================== */
/**
  * Turn head left and right around screen vertical axis.
  */
static void ViewWlkRotV( V3dState *view, double incr )
{
/* Set the axis of rotation: eye V axis.
 */
 view->rot_axis = v3d.v;

/* Set center of rotation at eye point.
 */
 view->center = view->eye;

/* Set the rotation angle based on given rotation increment.
 */
 view->angle = incr;

/* Rotate reference point about eye around found 
 * rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* ========================  ViewWlkRotW  ====================== */
/**
  * Spin head around screen out-of-plane axis.
  */
static void ViewWlkRotW( V3dState *view, double incr )
{
/* Set the axis of rotation: eye W axis.
 */
 view->rot_axis = v3d.w;

/* Set center of rotation at eye point.
 */
 view->center = view->eye;

/* Set the rotation angle based on given rotation increment.
 */
 view->angle = incr;

/* Rotate reference point about eye around found 
 * rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* ========================  ViewWlkTranU  ===================== */
/**
 * Move observer left and right along screen horizontal axis.
 */
static void ViewWlkTranU( V3dState *view, double incr )
{
 double     min_size;          /* minimum window size            */
 double     tran_incr;         /* translation increment          */

/* Obtain translation increment in eye system at projection plane.
 */
 min_size = MIN( view->right - view->left, view->top - view->bottom );
 tran_incr = min_size * incr;

/* Translate reference and eye points in eye U direction.
 */
 view->transl.x = tran_incr * v3d.u.x;
 view->transl.y = tran_incr * v3d.u.y;
 view->transl.z = tran_incr * v3d.u.z;
 ViewTranslate( view );
}

/* ========================  ViewWlkTranV  ===================== */
/**
 * Move observer down and up along screen vertical axis.
 */
static void ViewWlkTranV( V3dState *view, double incr )
{
 double     min_size;          /* minimum window size            */
 double     tran_incr;         /* translation increment          */

/* Obtain translation increment in eye system at projection plane.
 */
 min_size = MIN( view->right - view->left, view->top - view->bottom );
 tran_incr = min_size * incr;

/* Translate reference and eye points in eye V direction.
 */
 view->transl.x = tran_incr * v3d.v.x;
 view->transl.y = tran_incr * v3d.v.y;
 view->transl.z = tran_incr * v3d.v.z;
 ViewTranslate( view );
}

/* ========================  ViewWlkTranW  ===================== */
/**
 * Move observer forward and back along screen out-of-plane axis.
 */
static void ViewWlkTranW( V3dState *view, double incr )
{
 double     min_size;          /* minimum window size            */
 double     tran_incr;         /* translation increment          */

/* Obtain translation increment in eye system at projection plane.
 * Foward movement is in -W direction.
 */
 min_size = MIN( view->right - view->left, view->top - view->bottom );
 tran_incr = -(min_size * incr);

/* Translate reference and eye points in eye W direction.
 */
 view->transl.x = tran_incr * v3d.w.x;
 view->transl.y = tran_incr * v3d.w.y;
 view->transl.z = tran_incr * v3d.w.z;
 ViewTranslate( view );
}

/* ========================  ViewObjRotX  ====================== */
/**
 * Rotate object around its X axis.
 */
static void ViewObjRotX( V3dState *view, double incr )
{
/* Set the axis of rotation: object X axis.
 */
 view->rot_axis.x = 1.0;
 view->rot_axis.y = 0.0;
 view->rot_axis.z = 0.0;

/* Compute center of bounding box.
 */
 view->center.x = (view->box.xmn + view->box.xmx) * 0.5;
 view->center.y = (view->box.ymn + view->box.ymx) * 0.5;
 view->center.z = (view->box.zmn + view->box.zmx) * 0.5;

/* Set the rotation angle based on given rotation increment.
 */
 view->angle = incr;

/* Rotate eye position and reference pt. about center of
 * bounding box around rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* ========================  ViewObjRotY  ====================== */
/**
 * Rotate object around its Y axis.
 */
static void ViewObjRotY( V3dState *view, double incr )
{
/* Set the axis of rotation: object Y axis.
 */
 view->rot_axis.x = 0.0;
 view->rot_axis.y = 1.0;
 view->rot_axis.z = 0.0;

/* Compute center of bounding box.
 */
 view->center.x = (view->box.xmn + view->box.xmx) * 0.5;
 view->center.y = (view->box.ymn + view->box.ymx) * 0.5;
 view->center.z = (view->box.zmn + view->box.zmx) * 0.5;

/* Set the rotation angle based on given rotation increment.
 */
 view->angle = incr;

/* Rotate eye position and reference point about center 
 * around rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* ========================  ViewObjRotZ  ====================== */
/**
 * Rotate object around its Z axis.
 */
static void ViewObjRotZ( V3dState *view, double incr )
{
/* Set the axis of rotation: object Z axis.
 */
 view->rot_axis.x = 0.0;
 view->rot_axis.y = 0.0;
 view->rot_axis.z = 1.0;

/* Compute center of bounding box.
 */
 view->center.x = (view->box.xmn + view->box.xmx) * 0.5;
 view->center.y = (view->box.ymn + view->box.ymx) * 0.5;
 view->center.z = (view->box.zmn + view->box.zmx) * 0.5;

/* Set the rotation angle based on given rotation increment.
 */
 view->angle = incr;

/* Rotate eye position and reference pt. about center of
 * bounding box around rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* ========================  ViewObjTranX  ===================== */
/**
 * Translate object along its X direction.
 */
static void ViewObjTranX( V3dState *view, double incr )
{
 double     min_size;          /* minimum window size            */
 double     tran_incr;         /* translation increment          */

/* Obtain mouse increment in eye system at projection plane.
 */
 min_size = MIN( view->right - view->left, view->top - view->bottom );
 tran_incr = min_size * incr;

/* Translate reference and eye points in object X direction.
 */
 view->transl.x = tran_incr;
 view->transl.y = 0.0;
 view->transl.z = 0.0;
 ViewTranslate( view );
}

/* ========================  ViewObjTranY  ===================== */
/**
 * Translate object along its Y direction.
 */
static void ViewObjTranY( V3dState *view, double incr )
{
 double     min_size;          /* minimum window size            */
 double     tran_incr;         /* translation increment          */

/* Obtain mouse increment in eye system at projection plane.
 */
 min_size = MIN( view->right - view->left, view->top - view->bottom );
 tran_incr = min_size * incr;

/* Translate reference and eye points in object Y direction.
 */
 view->transl.x = 0.0;
 view->transl.y = tran_incr;
 view->transl.z = 0.0;
 ViewTranslate( view );
}

/* ========================  ViewObjTranZ  ===================== */
/**
 * Translate object along its Z direction.
 */
static void ViewObjTranZ( V3dState *view, double incr )
{
 double     min_size;          /* minimum window size            */
 double     tran_incr;         /* translation increment          */

/* Obtain mouse increment in eye system at projection plane.
 */
 min_size = MIN( view->right - view->left, view->top - view->bottom );
 tran_incr = min_size * incr;

/* Translate reference and eye points in object Z direction.
 */
 view->transl.x = 0.0;
 view->transl.y = 0.0;
 view->transl.z = tran_incr;
 ViewTranslate( view );
}

/* ====================  SetupMouseControl  =================== */

static void SetupMouseControl( V3dState *view, int b, int crtl, int shift )
{
 view->update_type = V3dNoUpdate;

/* If visualization model is object-based, just return.
 * (no mouse control).
 */
 if( view->model == 'o' )
  return;
 
/* If CRTL key is pressed with left mouse button pressed,
 * scale view window.
 * Else if CRTL key is pressed with right mouse button pressed,
 * change eye distance to projection plane.
 * Else if SHIFT key is pressed with left mouse butto pressed,
 * pan view window.
 */
 if( crtl )
 {
  if( b == IUP_BUTTON1 )
   view->update_type = V3dScaleWindow;
  else if( b == IUP_BUTTON3 )
   view->update_type = V3dEyeDistance;
  return;
 }
 else if( shift )
 {
  if( b == IUP_BUTTON1 )
   view->update_type = V3dPanWindow;
  return;
 }

/* Set up mouse control based on current model (screen-based or walk-through),
 * on current control mechanism (rotation or translation), and on which
 * button was pressed.
 */
 switch( view->model )
 {
  case 's':  /* screen-based model */
  switch( view->control )
  {
   case 'r':  /* rotation control */
    if( b == IUP_BUTTON1 )
     view->update_type = V3dScrRotate;
    else if( b == IUP_BUTTON3 )
     view->update_type = V3dScrSpin;
   break;
   case 't':  /* translation control */
    if( b == IUP_BUTTON1 )
     view->update_type = V3dScrTranslateUV;
    else if( b == IUP_BUTTON3 )
     view->update_type = V3dScrTranslateW;
   break;
  }
  break;

  case 'w':  /* walk-through model */
  switch( view->control )
  {
   case 'r':  /* rotation control */
    if( b == IUP_BUTTON1 )
     view->update_type = V3dWlkRotate;
    else if( b == IUP_BUTTON3 )
     view->update_type = V3dWlkSpin;
   break;
   case 't':  /* translation control */
    if( b == IUP_BUTTON1 )
     view->update_type = V3dWlkTranslateUV;
    else if( b == IUP_BUTTON3 )
     view->update_type = V3dWlkTranslateW;
   break;
  }
  break;
 }
}

/* ====================  ViewUpdateMouseMv  ==================== */

static void ViewUpdateMouseMv( V3dState *view, double delta_x, double delta_y )
{
/* Compute eye local system axes.
 */
 BuildEyeSystem( view );

 switch( view->update_type )
 {
  case V3dScaleWindow:
   ViewScaleMouse( view, delta_x, delta_y );
   break;
  case V3dPanWindow:
   ViewPanMouse( view, delta_x, delta_y );
   break;
  case V3dEyeDistance:
   ViewEyeDistMouse( view, delta_x, delta_y );
   break;
  case V3dScrRotate:
   ViewScrRotMouse( view, delta_x, delta_y );
   break;
  case V3dScrSpin:
   ViewScrSpinMouse( view, delta_x, delta_y );
   break;
  case V3dScrTranslateUV:
   ViewScrTranUVMouse( view, delta_x, delta_y );
   break;
  case V3dScrTranslateW:
   ViewScrTranWMouse( view, delta_x, delta_y );
   break;
  case V3dWlkRotate:
   ViewWlkRotMouse( view, delta_x, delta_y );
   break;
  case V3dWlkSpin:
   ViewWlkSpinMouse( view, delta_x, delta_y );
   break;
  case V3dWlkTranslateUV:
   ViewWlkTranUVMouse( view, delta_x, delta_y );
   break;
  case V3dWlkTranslateW:
   ViewWlkTranWMouse( view, delta_x, delta_y );
   break;
  case V3dNoUpdate:
  default:
   break;
 }
}

/* ======================  ViewScaleMouse  ===================== */

static void ViewScaleMouse( V3dState *view, double delta_x, double delta_y )
{
 double     midx,midy;         /* center of view volume          */
 double     width,height;      /* view volume lateral sizes      */
 double     scalefac;

/* Compute window center and sizes (in eye system at projection plane).
 */
 midx   = (view->right + view->left) * 0.5;
 midy   = (view->top + view->bottom) * 0.5;
 width  = (view->right - view->left) * 0.5;
 height = (view->top - view->bottom) * 0.5;

/* Scale view volume according to mouse increment.
 */
 scalefac = 1.0 + ScaleFactor * delta_x;
 view->left   = midx - (width / scalefac);
 view->right  = midx + (width / scalefac);
 view->bottom = midy - (height / scalefac);
 view->top    = midy + (height / scalefac);
}

/* =======================  ViewPanMouse  ====================== */

static void ViewPanMouse( V3dState *view, double delta_x, double delta_y )
{
/* Obtain mouse increment in eye system at projection plane.
 */
 delta_x *= view->right - view->left;
 delta_y *= view->top - view->bottom;

/* Pan lateral view window based on mouse increment.
 */
 view->right  -= delta_x;
 view->left   -= delta_x;
 view->bottom -= delta_y;
 view->top    -= delta_y;
}

/* =====================  ViewEyeDistMouse  ==================== */

static void ViewEyeDistMouse( V3dState *view, double delta_x, double delta_y )
{
/* Obtain mouse increment in eye system at projection plane.
 */
 delta_x *= view->right - view->left;

/* Avoid moving the eye inside the visualization volume.
 */
 if( delta_x >= view->front )
  return;

/* Translate eye point in the z local direction.
 */
 view->eye.x -= delta_x * v3d.w.x;
 view->eye.y -= delta_x * v3d.w.y;
 view->eye.z -= delta_x * v3d.w.z;

/* Keep front and back planes in previous absolute positions.
 */
 view->front -= delta_x;
 view->back  -= delta_x;
}

/* ======================  ViewScrRotMouse  ==================== */
/**
 * Turn object around an axis on the screen perpendicular to 
 * the mouse movement.
 */
static void ViewScrRotMouse( V3dState *view, double delta_x, double delta_y )
{
/* Find the axis of rotation based on the U and V eye axes 
 * and on the increment of mouse position.  Rotation axis
 * lies on the plane formed by U and V, and is perpendicular
 * to the vector of increment of mouse position on that plane
 * (rot_axis = - delta_y * U + delta_x * V).
 */
/*** COMPLETE HERE: 15 ***/
 view->rot_axis.x = 
 view->rot_axis.y = 
 view->rot_axis.z = 

/* Set center of rotation at reference point.
 */
/*** COMPLETE HERE: 16 ***/
 view->center = 
 
/* Compute the rotation angle based on the increment of
 * mouse position.
 * Screen-based ==> rotate eye contrary to target object rotation.
 */
/*** COMPLETE HERE: 17 ***/
 view->angle = 
 
/* Rotate eye position about reference point around found 
 * rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* =====================  ViewScrSpinMouse  ==================== */
/**
 * Spin object around the screen out-of-plane axis using the 
 * mouse position to compute the spin angle.
 */
static void ViewScrSpinMouse( V3dState *view, double delta_x, double delta_y )
{
 double     x, y;              /* current mouse position             */
 double     cx, cy;            /* center of rotation in raster space */

 x = (double)v3d.prevpt_x + delta_x * (double)view->hsize;
 y = (double)v3d.prevpt_y + delta_y * (double)view->vsize;
 cx = (double)view->hsize * 0.5;
 cy = (double)view->vsize * 0.5;

/* Set the axis of rotation: eye W axis.
 */
/*** COMPLETE HERE: 18 ***/
 view->rot_axis = 
 
/* Set center of rotation at reference point.
 */
/*** COMPLETE HERE: 19 ***/
 view->center = 
 
/* Compute the spin angle based on the increment of mouse position.
 * Screen-based ==> rotate eye contrary to target object rotation.
 */
 view->angle = atan2( y-cy, x-cx ) - 
               atan2( (double)v3d.prevpt_y-cy, (double)v3d.prevpt_x-cx );
 view->angle = -deg( view->angle );

/* Rotate view up-vector around the view axis.
 */
 ViewRotate( view );
}

/* =====================  ViewScrTranUVMouse  ================== */
/**
 * Move object along an axis on the screen in the direction of
 * the mouse movement.
 */
static void ViewScrTranUVMouse( V3dState *view, double delta_x, double delta_y )
{
 double     tranu_incr;        /* U translation increment        */
 double     tranv_incr;        /* V translation increment        */

/* Obtain mouse increment in eye system at projection plane.
 * Screen-based ==> move eye and ref. contrary to target object translation.
 */
/*** COMPLETE HERE: 20 ***/
 tranu_incr = 
 tranv_incr = 

/* Translate reference and eye points in eye U and V directions.
 */
/*** COMPLETE HERE: 21 ***/
 view->transl.x = 
 view->transl.y = 
 view->transl.z = 
 ViewTranslate( view );
}

/* =====================  ViewScrTranWMouse  =================== */
/**
 * Move object along the out-plane screen axis using the mouse
 * movement to compute the translation increment.
 */
static void ViewScrTranWMouse( V3dState *view, double delta_x, double delta_y )
{
 double     tranw_incr;        /* W translation increment        */

/* Obtain mouse increment in eye system at projection plane.
 * Screen-based ==> move eye and ref. contrary to target object translation.
 * Foward movement is in -W direction (sign changes twice).
 */
 tranw_incr = delta_y * (view->top - view->bottom);

/* Translate reference and eye points in eye W direction.
 */
 view->transl.x = tranw_incr * v3d.w.x;
 view->transl.y = tranw_incr * v3d.w.y;
 view->transl.z = tranw_incr * v3d.w.z;
 ViewTranslate( view );
}

/* ======================  ViewWlkRotMouse  ==================== */
/**
 * Turn head around an axis on the screen perpendicular to 
 * the mouse movement.
 */
static void ViewWlkRotMouse( V3dState *view, double delta_x, double delta_y )
{
/* Find the axis of rotation based on the U and V eye axes 
 * and on the increment of mouse position.  Rotation axis
 * lies on the plane formed by U and V, and in perpendicular
 * to the vector of increment of mouse position on that plane
 * (rot_axis = - delta_y * U + delta_x * V).
 */
 view->rot_axis.x = delta_x * v3d.v.x - delta_y * v3d.u.x;
 view->rot_axis.y = delta_x * v3d.v.y - delta_y * v3d.u.y;
 view->rot_axis.z = delta_x * v3d.v.z - delta_y * v3d.u.z;

/* Set center of rotation at eye point.
 */
 view->center = view->eye;

/* Compute the rotation angle based on the increment of
 * mouse position.
 */
 view->angle = -Len2( delta_x, delta_y ) * 180.0;

/* Rotate reference point about eye around found 
 * rotation axis.
 * Rotate view up-vector around rotation axis.
 */
 ViewRotate( view );
}

/* =====================  ViewWlkSpinMouse  ==================== */
/**
 * Spin head around the screen out-of-plane axis using the 
 * mouse position to compute the spin angle.
 */
static void ViewWlkSpinMouse( V3dState *view, double delta_x, double delta_y )
{
 double     x, y;              /* current mouse position             */
 double     cx, cy;            /* center of rotation in raster space */

 x = (double)v3d.prevpt_x + delta_x * (double)view->hsize;
 y = (double)v3d.prevpt_y + delta_y * (double)view->vsize;
 cx = (double)view->hsize * 0.5;
 cy = (double)view->vsize * 0.5;

/* Set the axis of rotation: eye W axis.
 */
 view->rot_axis = v3d.w;

/* Set center of rotation at eye point.
 */
 view->center = view->eye;

/* Compute the spin angle based on the increment of
 * mouse position.
 */
 view->angle = atan2( y-cy, x-cx ) - 
               atan2( (double)v3d.prevpt_y-cy, (double)v3d.prevpt_x-cx );
 view->angle = deg( view->angle );

/* Rotate view up-vector around the view axis.
 */
 ViewRotate( view );
}

/* =====================  ViewWlkTranUVMouse  ================== */
/**
 * Move observer along an axis on the screen in the direction of
 * the mouse movement.
 */
static void ViewWlkTranUVMouse( V3dState *view, double delta_x, double delta_y )
{ 
 double     tranu_incr;        /* U translation increment        */
 double     tranv_incr;        /* V translation increment        */

/* Obtain mouse increment in eye system at projection plane.
 */
 tranu_incr = delta_x * (view->right - view->left);
 tranv_incr = delta_y * (view->top - view->bottom);

/* Translate reference and eye points in eye U and V directions.
 */
 view->transl.x = tranu_incr * v3d.u.x + tranv_incr * v3d.v.x;
 view->transl.y = tranu_incr * v3d.u.y + tranv_incr * v3d.v.y;
 view->transl.z = tranu_incr * v3d.u.z + tranv_incr * v3d.v.z;
 ViewTranslate( view );
}

/* ====================  ViewWlkTranWMouse  =================== */
/**
 * Move observer along the out-plane screen axis using the mouse
 * movement to compute the translation increment.
 */
static void ViewWlkTranWMouse( V3dState *view, double delta_x, double delta_y )
{
 double     tranw_incr;        /* W translation increment        */

/* Obtain mouse increment in eye system at projection plane.
 * Foward movement is in -W direction.
 */
 tranw_incr = - delta_y * (view->top - view->bottom);

/* Translate reference and eye points in eye W direction.
 */
 view->transl.x = tranw_incr * v3d.w.x;
 view->transl.y = tranw_incr * v3d.w.y;
 view->transl.z = tranw_incr * v3d.w.z;
 ViewTranslate( view );
}

/* =========================  ProjectBox  ====================== */

static void ProjectBox( V3dState *view, double *left, double *right,
                                        double *bottom, double *top )
{
 Point  ref;               /* reference point transformed to screen system */
 Point  box[8];            /* model box transformed to screen system */
 double pxmin, pxmax, pymin, pymax;   /* projected bounding box */
 double sizex, sizey;

 ref = view->ref;
 box[0].x = view->box.xmn;  box[0].y = view->box.ymn;  box[0].z = view->box.zmn;
 box[1].x = view->box.xmx;  box[1].y = view->box.ymn;  box[1].z = view->box.zmn;
 box[2].x = view->box.xmn;  box[2].y = view->box.ymx;  box[2].z = view->box.zmn;
 box[3].x = view->box.xmx;  box[3].y = view->box.ymx;  box[3].z = view->box.zmn;
 box[4].x = view->box.xmn;  box[4].y = view->box.ymn;  box[4].z = view->box.zmx;
 box[5].x = view->box.xmx;  box[5].y = view->box.ymn;  box[5].z = view->box.zmx;
 box[6].x = view->box.xmn;  box[6].y = view->box.ymx;  box[6].z = view->box.zmx;
 box[7].x = view->box.xmx;  box[7].y = view->box.ymx;  box[7].z = view->box.zmx;

/* Create a transformation matrix for the current camera transformation 
 * that includes the perpective distortion.  If current visualization
 * type is perspective, projection plane is considered at the front 
 * clip plane.
 */
 t3dIdentity( );
/*** COMPLETE HERE: 22 ***/

 if( view->type == 'p' )
 {
/*** COMPLETE HERE: 23 ***/
 }
 
/* Transform reference point and 3D bounding box points to screen 
 * coordinates (eye system with distortion holding depth information).
 */
 t3dTransform( &ref.x, &ref.y, &ref.z );
 t3dTransform( &box[0].x, &box[0].y, &box[0].z );
 t3dTransform( &box[1].x, &box[1].y, &box[1].z );
 t3dTransform( &box[2].x, &box[2].y, &box[2].z );
 t3dTransform( &box[3].x, &box[3].y, &box[3].z );
 t3dTransform( &box[4].x, &box[4].y, &box[4].z );
 t3dTransform( &box[5].x, &box[5].y, &box[5].z );
 t3dTransform( &box[6].x, &box[6].y, &box[6].z );
 t3dTransform( &box[7].x, &box[7].y, &box[7].z );
 
/* Disregard screen z coordinate and get maximun and minimum
 * of projected box lateral coordinates.
 */
 pxmin = box[0].x;
 pxmin = MIN( pxmin, box[1].x );
 pxmin = MIN( pxmin, box[2].x );
 pxmin = MIN( pxmin, box[3].x );
 pxmin = MIN( pxmin, box[4].x );
 pxmin = MIN( pxmin, box[5].x );
 pxmin = MIN( pxmin, box[6].x );
 pxmin = MIN( pxmin, box[7].x );
 pxmax = box[0].x;
 pxmax = MAX( pxmax, box[1].x );
 pxmax = MAX( pxmax, box[2].x );
 pxmax = MAX( pxmax, box[3].x );
 pxmax = MAX( pxmax, box[4].x );
 pxmax = MAX( pxmax, box[5].x );
 pxmax = MAX( pxmax, box[6].x );
 pxmax = MAX( pxmax, box[7].x );
 pymin = box[0].y;
 pymin = MIN( pymin, box[1].y );
 pymin = MIN( pymin, box[2].y );
 pymin = MIN( pymin, box[3].y );
 pymin = MIN( pymin, box[4].y );
 pymin = MIN( pymin, box[5].y );
 pymin = MIN( pymin, box[6].y );
 pymin = MIN( pymin, box[7].y );
 pymax = box[0].y;
 pymax = MAX( pymax, box[1].y );
 pymax = MAX( pymax, box[2].y );
 pymax = MAX( pymax, box[3].y );
 pymax = MAX( pymax, box[4].y );
 pymax = MAX( pymax, box[5].y );
 pymax = MAX( pymax, box[6].y );
 pymax = MAX( pymax, box[7].y );

/* Compute the maximum distances in x and y from projected
 * reference point to projected bounding box.
 */
 sizex = MAX( ABS( pxmax - ref.x ), ABS( ref.x - pxmin ) );
 sizey = MAX( ABS( pymax - ref.y ), ABS( ref.y - pymin ) );

/* Compute view window limits such that reference point is
 * the window center.
 */
 *left   = ref.x - sizex;
 *right  = ref.x + sizex;
 *bottom = ref.y - sizey;
 *top    = ref.y + sizey;
}


/*
** ---------------------------------------------------------------
** Public functions:
*/

/* ==========================  v3dOpen  ======================= */
 
int v3dOpen( void *canvas )
{
 V3dState  *view;

 view = GetViewState( canvas );
 if( view != NULL )
  v3dClose( canvas );

 view = (V3dState *)malloc( sizeof( V3dState ) );
 if( view == NULL )
  return( 0 );

 view->next = v3d.view_head;
 v3d.view_head = view;
 view->canvas = canvas;
 view->update_type = V3dNoUpdate;

/* Setup default model box.
 */
 view->box.xmn = 0.0;
 view->box.xmx = 1.0;
 view->box.ymn = 0.0;
 view->box.ymx = 1.0;
 view->box.zmn = 0.0;
 view->box.zmx = 1.0;

/* Setup default viewing parameters:
 * view up-vector in the object y direction;
 * perspective projection; screen-base viewing 
 * model; and rotation control.
 */
 view->defvup  = 'y';
 view->type    = 'p';
 view->model   = 's';
 view->control = 'r';

/* Setup default canvas sizes.
 */
 view->hsize = 100;
 view->vsize = 100;

/* Set up default visualization parameters.
 */
 view->eye.x    = 0.5;
 view->eye.y    = 0.5;
 view->eye.z    = 4.0;
 view->ref.x    = 0.5;
 view->ref.y    = 0.5;
 view->ref.z    = 0.5;
 view->vup.x    = 0.0;
 view->vup.y    = 1.0;
 view->vup.z    = 0.0;
 view->left     = 0.0;
 view->right    = 1.0;
 view->bottom   = 0.0;
 view->top      = 1.0;
 view->front    = 0.0;
 view->back     = 1.0;

 view->rot_axis.x = 0.0;
 view->rot_axis.y = 0.0;
 view->rot_axis.z = 1.0;
 view->angle = 0.0;
 view->center.x = 0.5;
 view->center.y = 0.5;
 view->center.z = 0.5;
 view->transl.x = 0.0;
 view->transl.y = 0.0;
 view->transl.z = 0.0;

/* Initialize saved viewing parameters.
 */
 view->saved      = 0;
 view->type_sav   = view->type;
 view->eye_sav    = view->eye;
 view->ref_sav    = view->ref;
 view->vup_sav    = view->vup;
 view->left_sav   = view->left;
 view->right_sav  = view->right;
 view->bottom_sav = view->bottom;
 view->top_sav    = view->top;
 view->front_sav  = view->front;
 view->back_sav   = view->back;

 return( 1 );
}

/* =========================  v3dClose  ======================= */
 
void v3dClose( void *canvas )
{
 V3dState  *view;
 V3dState  *prev;

 view = GetViewState( canvas );
 if( view == NULL )
  return;

 if( v3d.view_head == view )
 {
  v3d.view_head = view->next;
 }
 else
 {
  prev = v3d.view_head;
  while( ( prev->next != view ) && ( prev->next != NULL ) )
  {
   prev = prev->next;
  }
 
  if( prev->next != NULL )
  {
   prev->next = view->next;
  }
 }

 free( view );
}

/* =======================  v3dSetModelBox  =================== */
 
void v3dSetModelBox( void *canvas, 
                     double xmin, double xmax,
                     double ymin, double ymax,
                     double zmin, double zmax )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->box.xmn = xmin;
 view->box.xmx = xmax;
 view->box.ymn = ymin;
 view->box.ymx = ymax;
 view->box.zmn = zmin;
 view->box.zmx = zmax;
}

/* =======================  v3dSetDefVup  ===================== */
 
void v3dSetDefVup( void *canvas, char defvup )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->defvup = defvup;

 view->update_type = V3dNoUpdate;
}
 
/* ========================  v3dSetType  ====================== */
 
void v3dSetType( void *canvas, char vtype )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->type = vtype;

 view->update_type = V3dNoUpdate;
}

/* =======================  v3dSetModel  ====================== */
 
void v3dSetModel( void *canvas, char vmodel )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->model = vmodel;

 view->update_type = V3dNoUpdate;
}

/* ======================  v3dSetControl  ===================== */
 
void v3dSetControl( void *canvas, char control )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->control = control;

 view->update_type = V3dNoUpdate;
}

/* ========================  v3dDefault  ====================== */
 
void v3dDefault( void *canvas )
{
 V3dState  *view;
 double    left, right,
           bottom, top;       /* projected bounding box         */
 double    maxsize;           /* maximum bounding box size      */
 double    width;             /* window width size              */
 double    height;            /* window height size             */
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

/* Compute default visualization volume and parameters 
 * based on object/environment dimensions.
 * Default reference point is at the center of the 
 * bounding box.  Default eye position is found in
 * accordance to the default view up-vector direction.
 */
 maxsize = MAX( view->box.xmx-view->box.xmn, view->box.ymx-view->box.ymn );
 maxsize = MAX( maxsize, view->box.zmx-view->box.zmn );

 view->ref.x = (view->box.xmn + view->box.xmx) * 0.5;
 view->ref.y = (view->box.ymn + view->box.ymx) * 0.5;
 view->ref.z = (view->box.zmn + view->box.zmx) * 0.5;

 switch( view->defvup )
 {
  case 'x':
   view->eye.x = view->ref.x;
   view->eye.y = view->ref.y + Sqrt3 * 1.5 * maxsize;
   view->eye.z = view->ref.z;
   view->vup.x = 1.0;
   view->vup.y = 0.0;
   view->vup.z = 0.0;
   break;
  case 'y':
   view->eye.x = view->ref.x;
   view->eye.y = view->ref.y;
   view->eye.z = view->ref.z + Sqrt3 * 1.5 * maxsize;
   view->vup.x = 0.0;
   view->vup.y = 1.0;
   view->vup.z = 0.0;
   break;
  case 'z':
  default:
   view->eye.x = view->ref.x + Sqrt3 * 1.5 * maxsize;
   view->eye.y = view->ref.y;
   view->eye.z = view->ref.z;
   view->vup.x = 0.0;
   view->vup.y = 0.0;
   view->vup.z = 1.0;
   break;
 }

 view->front = Sqrt3 * maxsize;
 view->back  = Sqrt3 * 2.0 * maxsize;

/* Compute default lateral visualization volume parameters 
 * based on dimensions of projected bounding box.
 */
 ProjectBox( view, &left, &right, &bottom, &top );
 width  = right - left;
 height = top - bottom;
 view->right  = FitFactor * 0.5 * width;
 view->left   = -view->right;
 view->top    = FitFactor * 0.5 * height;
 view->bottom = -view->top;
 AdjustWindow( view, &view->left, &view->right, &view->bottom, &view->top );

/* If viewing parameters were never stored, save default viewing 
 * parameters for restoring later.
 */
 if( !view->saved )
 {
  view->type_sav   = view->type;
  view->eye_sav    = view->eye;
  view->ref_sav    = view->ref;
  view->vup_sav    = view->vup;
  view->left_sav   = view->left;
  view->right_sav  = view->right;
  view->bottom_sav = view->bottom;
  view->top_sav    = view->top;
  view->front_sav  = view->front;
  view->back_sav   = view->back;
 }
}

/* =========================  v3dResize  ====================== */
 
void v3dResize( void *canvas, int width, int height )
{
 V3dState  *view;

 view = GetViewState( canvas );
 if( view == NULL )
  return;

/* Update canvas sizes in pixels and set flag for doing resize.
 */
 view->hsize = width;
 view->vsize = height;

/* Adjust window to current viewport ratio.
 */
 AdjustWindow( view, &view->left, &view->right, &view->bottom, &view->top );
}

/* =========================  v3dZoomIn  ====================== */

static void _v3dZoomInOutGen( void *canvas, int in );

void v3dZoomIn( void *canvas )
{
 _v3dZoomInOutGen( canvas, 1 );
}

/* =========================  v3dZoomOut  ===================== */

void v3dZoomOut( void *canvas )
{
 _v3dZoomInOutGen( canvas, 0 );
}

/* ----------------------  v3dZoomInOutGen  ------------------- */

static void _v3dZoomInOutGen( void *canvas, int in )
{
 V3dState  *view;
 double    center_w, center_h;       /* new center of view volume */
 double    width, height;            /* view volume lateral sizes */

 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

/* Adjust window based on the current zoom factor.
 * The center of the new window is maintained.
 */
 center_w = (view->right + view->left) * 0.5;
 center_h = (view->top + view->bottom) * 0.5;
 width = (view->right - view->left) * 0.5;
 height = (view->top - view->bottom) * 0.5;
 if( in )
 {
  width /= ZoomFactor;
  height /= ZoomFactor;
 }
 else
 {
  width *= ZoomFactor;
  height *= ZoomFactor;
 }
 view->left   = center_w - width;
 view->right  = center_w + width;
 view->bottom = center_h - height;
 view->top    = center_h + height;
}


/* ==========================  v3dFit  ======================== */
 
void v3dFit( void *canvas )
{
 V3dState  *view;
 double    left, right,
           bottom, top;       /* projected bounding box         */
 double    cx, cy;            /* window center                  */
 double    width;             /* window width size              */
 double    height;            /* window height size             */
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

/* Compute lateral visualization volume parameters 
 * based on dimensions of projected bounding box and 
 * on current fit factor.
 */
 ProjectBox( view, &left, &right, &bottom, &top );
 cx = (left + right) * 0.5;
 cy = (bottom + top) * 0.5;
 width  = right - left;
 height = top - bottom;
 view->left   = cx - FitFactor * 0.5 * width;
 view->right  = cx + FitFactor * 0.5 * width;
 view->bottom = cy - FitFactor * 0.5 * height;
 view->top    = cy + FitFactor * 0.5 * height;
 AdjustWindow( view, &view->left, &view->right, &view->bottom, &view->top );
}
 
/* =======================  v3dProjPlane  ===================== */

void v3dProjPlane( void *canvas, char *plane )
{
 V3dState  *view;
 double    d;                 /* distance from eye to ref. pt.  */
 Vector    w;                 /* target eye system z direction  */
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

 d = Len3( view->eye.x-view->ref.x, 
           view->eye.y-view->ref.y, 
           view->eye.z-view->ref.z );

/* Set target view direction based on given projection plane.
 */
 if(      ( !strcmp( plane, "xy"   ) ) || 
          ( !strcmp( plane, "XY"   ) ) ||
          ( !strcmp( plane, "+x+y" ) ) ||
          ( !strcmp( plane, "+X+Y" ) ) )
 {
  w.x = 0.0;  w.y = 0.0;  w.z = 1.0;
  view->vup.x = 0.0;  view->vup.y = 1.0;  view->vup.z = 0.0;
 }
 else if( ( !strcmp( plane, "xz"   ) ) || 
          ( !strcmp( plane, "XZ"   ) ) ||
          ( !strcmp( plane, "+x+z" ) ) ||
          ( !strcmp( plane, "+X+Z" ) ) )
 {
  w.x = 0.0;  w.y = -1.0;  w.z = 0.0;
  view->vup.x = 0.0;  view->vup.y = 0.0;  view->vup.z = 1.0;
 }
 else if( ( !strcmp( plane, "yz"   ) ) || 
          ( !strcmp( plane, "YZ"   ) ) ||
          ( !strcmp( plane, "+y+z" ) ) ||
          ( !strcmp( plane, "+Y+Z" ) ) )
 {
  w.x = 1.0;  w.y = 0.0;  w.z = 0.0;
  view->vup.x = 0.0;  view->vup.y = 0.0;  view->vup.z = 1.0;
 }
 else if( ( !strcmp( plane, "IsoXYZ"    ) ) || 
          ( !strcmp( plane, "Iso+X+Y+Z" ) ) ||
          ( !strcmp( plane, "Iso+x+y+z" ) ) )
 {
  w.x = 1.0/Sqrt3;
  w.y = 1.0/Sqrt3;
  w.z = 1.0/Sqrt3;
  view->vup.x = -1.0/Sqrt6;
  view->vup.y = -1.0/Sqrt6;
  view->vup.z =  2.0/Sqrt6;
 }
 else if( ( !strcmp( plane, "IsoZXY"    ) ) || 
          ( !strcmp( plane, "Iso+Z+X+Y" ) ) ||
          ( !strcmp( plane, "Iso+z+x+y" ) ) )
 {
  w.x = 1.0/Sqrt3;
  w.y = 1.0/Sqrt3;
  w.z = 1.0/Sqrt3;
  view->vup.x = -1.0/Sqrt6;
  view->vup.y =  2.0/Sqrt6;
  view->vup.z = -1.0/Sqrt6;
 }
 else if( ( !strcmp( plane, "IsoYZX"    ) ) || 
          ( !strcmp( plane, "Iso+Y+Z+X" ) ) ||
          ( !strcmp( plane, "Iso+y+z+x" ) ) )
 {
  w.x = 1.0/Sqrt3;
  w.y = 1.0/Sqrt3;
  w.z = 1.0/Sqrt3;
  view->vup.x =  2.0/Sqrt6;
  view->vup.y = -1.0/Sqrt6;
  view->vup.z = -1.0/Sqrt6;
 }
 else
 {
  return;
 }

/* Set eye position such that its distance to the reference
 * point is maintained and its position is found in 
 * accordance to the specified view direction.
 */
 view->eye.x = view->ref.x + d * w.x;
 view->eye.y = view->ref.y + d * w.y;
 view->eye.z = view->ref.z + d * w.z;
}

/* ========================  v3dMouseBt  ====================== */

void v3dMouseBt( void *canvas, int b, int mode, int x, int y, char *mousecfg )
{
 V3dState  *view;

/* Switch to the viewing spec. corresponding to given canvas.
 */
 view = GetViewState( canvas );
 if( view == NULL )
  return;

/* Change from IUP top-left raster coordinate origin to bottom-left origin.
 */
 y = (view->vsize - 1) - y;

 if( mode == ButtonReleased )
 {
/* If button pressed, set up change view mode.
 */
  v3d.change_view = 1;
  view->update_type = V3dNoUpdate;
 
/* Set up view package to next mouse movement.
 */
  SetupMouseControl( view, b, 
                     ControlPressed( mousecfg ), ShiftPressed( mousecfg ) );

/* Set up initial mouse point and previous mouse point to given point.
 */
  v3d.setuppt_x = v3d.prevpt_x = x;
  v3d.setuppt_y = v3d.prevpt_y = y;
 }
 else
 {
/* If button released, close change view mode.
 */
  v3d.change_view = 0;
 }
}

/* ========================  v3dMouseMv  ======================= */

void v3dMouseMv( void *canvas, int x, int y, char *mousecfg )
{
 V3dState  *view;
 double    delta_x, delta_y; /* normalized mouse movement        */

/* Reject point if not changing view.
 */
 if( ! v3d.change_view )
  return;

/* Switch to the viewing spec. corresponding to given canvas.
 */
 view = GetViewState( canvas );
 if( view == NULL )
  return;

/* Change from IUP top-left raster coordinate origin to bottom-left origin.
 */
 y = (view->vsize - 1) - y;

/* Check to see whether current mouse point is too close to the previous 
 * mouse position point.  If that is the case, disregard point and return.
 */
 if( (ABS(x - v3d.prevpt_x) < 4) && (ABS(y - v3d.prevpt_y) < 4) )
  return;

/* Compute normalized mouse movement.
 */
 delta_x = ((double)x - (double)v3d.prevpt_x) / (double)view->hsize;
 delta_y = ((double)y - (double)v3d.prevpt_y) / (double)view->vsize;

/* Change view according to current mouse position.
 */
 ViewUpdateMouseMv( view, delta_x, delta_y );

/* Update previous mouse point to last point.
 */
 v3d.prevpt_x = x;
 v3d.prevpt_y = y;
}

/* =======================  v3dIncrCrtl  ====================== */
 
void v3dIncrCrtl( void *canvas, char axis, char dir )
{
 V3dState  *view;

 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

/* Compute eye local system axes.
 */
 BuildEyeSystem( view );

/* Depending on the type of view model (screen based,
 * walk-through, or object based) use specific function
 * to manipulate model visualization on screen.
 */
 switch( view->model )
 {
  default:
  case 's':  /* screen-based model */
   if( view->control == 'r' )
   {
    switch( axis )
    {
     case 'x':
      if( dir == '-' )   /* -x -> turn object left (around eye -V axis) */
       ViewScrRotV( view, -RotFactor );
      else               /* +x -> turn object right (around eye +V axis) */
       ViewScrRotV( view, RotFactor );
      break;
     case 'y':
      if( dir == '-' )   /* -y -> turn object down (around eye +U axis) */
       ViewScrRotU( view, RotFactor );
      else               /* +y -> turn object up (around eye -U axis) */
       ViewScrRotU( view, -RotFactor );
      break;
     case 'z':
      if( dir == '-' )   /* -z -> spin object around eye +W axis */
       ViewScrRotW( view, RotFactor );
      else               /* +z -> spin object around eye -W axis */
       ViewScrRotW( view, -RotFactor );
      break;
     default:
      return;
    }
   }
   else if( view->control == 't' )
   {
    switch( axis )
    {
     case 'x':
      if( dir == '-' )   /* -x -> move object left (in eye -U direction) */
       ViewScrTranU( view, -DefaultIncr );
      else               /* +x -> move obj. right (in eye +U direction) */
       ViewScrTranU( view, DefaultIncr );
      break;
     case 'y':
      if( dir == '-' )   /* -y -> move object down (in eye -V direction) */
       ViewScrTranV( view, -DefaultIncr );
      else               /* +y -> move object up (in eye +V direction) */
       ViewScrTranV( view, DefaultIncr );
      break;
     case 'z':
      if( dir == '-' )   /* -z -> move object forward (in -W direction) */
       ViewScrTranW( view, -DefaultIncr );
      else               /* +z -> move object back (in +W direction) */
       ViewScrTranW( view, DefaultIncr );
      break;
     default:
      return;
    }
   break;
  case 'w':  /* walk-through model */
   if( view->control == 'r' )
   {
    switch( axis )
    {
     case 'x':
      if( dir == '-' )   /* -x -> turn head left (around eye +V axis ) */
       ViewWlkRotV( view, RotFactor );
      else               /* +x -> turn head right (around eye -V axis ) */
       ViewWlkRotV( view, -RotFactor );
      break;
     case 'y':
      if( dir == '-' )   /* -y -> turn head down (around eye -U axis ) */
       ViewWlkRotU( view, -RotFactor );
      else               /* +y -> turn head up (around eye +U axis ) */
       ViewWlkRotU( view, RotFactor );
      break;
     case 'z':
      if( dir == '-' )   /* -z -> spin head around eye +W axis  */
       ViewWlkRotW( view, RotFactor );
      else               /* +z -> spin head around eye -W axis  */
       ViewWlkRotW( view, -RotFactor );
      break;
     default:
      return;
    }
   }
   else if( view->control == 't' )
   {
    switch( axis )
    {
     case 'x':
      if( dir == '-' )   /* -x -> move observer left (in eye -U direction) */
       ViewWlkTranU( view, -DefaultIncr );
      else               /* +x -> move observer right (in eye +U direction) */
       ViewWlkTranU( view, DefaultIncr );
      break;
     case 'y':
      if( dir == '-' )   /* -y -> move observer down (in eye -V direction) */
       ViewWlkTranV( view, -DefaultIncr );
      else               /* +y -> move observer up (in eye +V direction) */
       ViewWlkTranV( view, DefaultIncr );
      break;
     case 'z':
      if( dir == '-' )   /* -z -> move observer forward (in eye -W direction) */
       ViewWlkTranW( view, -DefaultIncr );
      else               /* +z -> move observer back (in eye +W direction) */
       ViewWlkTranW( view, DefaultIncr );
      break;
     default:
      return;
    }
   }
   break;
  case 'o':  /* object-based model */
   if( view->control == 'r' )
   {
    switch( axis )
    {
     case 'x':
      if( dir == '-' )   /* -x -> rotate object around -X axis about center */
       ViewObjRotX( view, -RotFactor );
      else               /* +x -> rotate object around +X axis about center */
       ViewObjRotX( view, RotFactor );
      break;
     case 'y':
      if( dir == '-' )   /* -y -> rotate object around -Y axis about center */
       ViewObjRotY( view, -RotFactor );
      else               /* +y -> rotate object around +Y axis about center */
       ViewObjRotY( view, RotFactor );
      break;
     case 'z':
      if( dir == '-' )   /* -z -> rotate object around -Z axis about center */
       ViewObjRotZ( view, -RotFactor );
      else               /* +z -> rotate object around +Z axis about center */
       ViewObjRotZ( view, RotFactor );
      break;
     default:
      return;
    }
   }
   else if( view->control == 't' )
   {
    switch( axis )
    {
     case 'x':
      if( dir == '-' )   /* -x -> translate object in -X direction */
       ViewObjTranX( view, -DefaultIncr );
      else               /* +x -> translate object in +X direction */
       ViewObjTranX( view, DefaultIncr );
      break;
     case 'y':
      if( dir == '-' )   /* -y -> translate object in -Y direction */
       ViewObjTranY( view, -DefaultIncr );
      else               /* +y -> translate object in +Y direction */
       ViewObjTranY( view, DefaultIncr );
      break;
     case 'z':
      if( dir == '-' )   /* -z -> translate object in -Z direction */
       ViewObjTranZ( view, -DefaultIncr );
      else               /* +z -> translate object in +Z direction */
       ViewObjTranZ( view, DefaultIncr );
      break;
     default:
      return;
    }
   }
   break;
  }
 }
}

/* ========================  v3dKeyCrtl  ====================== */
 
void v3dKeyCrtl( void *canvas, int key )
{
 V3dState  *view;

 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

/* Handle general key control.
 */
 switch( key )
 {
  case K_D:   /* reset */
   v3dDefault( canvas );
   return;
  case K_F:   /* fit */
   v3dFit( canvas );
   return;
  case K_cP:   /* switch type to perspective */
   v3dSetType( canvas, 'p' );
   return;
  case K_cO:   /* switch type to orthographic */
   v3dSetType( canvas, 'o' );
   return;
  case K_R:   /* rotation control */
   v3dSetControl( canvas, 'r' );
   return;
  case K_T:   /* translation control */
   v3dSetControl( canvas, 't' );
   return;
  case K_S:   /* screen-based model */
   v3dSetModel( canvas, 's' );
   return;
  case K_W:   /* walk-though model */
   v3dSetModel( canvas, 'w' );
   return;
  case K_O:   /* object-based model */
   v3dSetModel( canvas, 'o' );
   return;
  case K_plus:  /* zoom-in */
   v3dZoomIn( canvas );
   return;
  case K_minus: /* zoom-out */
   v3dZoomOut( canvas );
   return;
 }

/* Compute eye local system axes.
 */
 BuildEyeSystem( view );

/* Depending on the type of view model (screen based,
 * walk-through, or object based) use specific function
 * to manipulate model visualization on screen.
 */
 switch( view->model )
 {
  default:
  case 's':  /* screen-based model */
   if( view->control == 'r' )
   {
    switch( key )
    {
     case K_LEFT:   /* Left -> turn object left (around eye -V axis) */
      ViewScrRotV( view, -RotFactor );
      break;
     case K_RIGHT:  /* Right -> turn object right (around eye +V axis) */
      ViewScrRotV( view, RotFactor );
      break;
     case K_DOWN:   /* Down -> turn object down (around eye +U axis) */
      ViewScrRotU( view, RotFactor );
      break;
     case K_UP:     /* Up -> turn object up (around eye -U axis) */
      ViewScrRotU( view, -RotFactor );
      break;
     case K_sLEFT:  /* Shift-Left -> spin object around eye +W axis */
      ViewScrRotW( view, RotFactor );
      break;
     case K_sRIGHT: /* Shift-Right -> spin object around eye -W axis */
      ViewScrRotW( view, -RotFactor );
      break;
     default:
      return;
    }
   }
   else if( view->control == 't' )
   {
    switch( key )
    {
     case K_LEFT:   /* Left -> move object left (in eye -U direction) */
      ViewScrTranU( view, -DefaultIncr );
      break;
     case K_RIGHT:  /* Right -> move object right (in eye +U direction) */
      ViewScrTranU( view, DefaultIncr );
      break;
     case K_DOWN:   /* Down -> move object down (in eye -V direction) */
      ViewScrTranV( view, -DefaultIncr );
      break;
     case K_UP:     /* Up -> move object up eye (in +V direction) */
      ViewScrTranV( view, DefaultIncr );
      break;
     case K_sDOWN:  /* Shift-Down -> move object back (in -W direction) */
      ViewScrTranW( view, -DefaultIncr );
      break;
     case K_sUP:    /* Shift-Up -> move object forward (in +W direction) */
      ViewScrTranW( view, DefaultIncr );
      break;
     default:
      return;
    }
   break;
  case 'w':  /* walk-through model (independ on the type of control) */
    switch( key )
    {
     case K_LEFT:    /* Left -> turn head left (around eye +V axis) */
      ViewWlkRotV( view, RotFactor );
      break;
     case K_RIGHT:   /* Right -> turn head right (around eye -V axis) */
      ViewWlkRotV( view, -RotFactor );
      break;
     case K_s:       /* s or */
     case K_DOWN:    /* Down -> move observer forward (in -W direction) */
      ViewWlkTranW( view, -DefaultIncr );
      break;
     case K_w:       /* w or */
     case K_UP:      /* Up -> move observer back (in +W direction) */
      ViewWlkTranW( view, DefaultIncr );
      break;
     case K_a:       /* a or */
     case K_sLEFT:   /* Shift-Left -> move observer left (in eye -U direct.) */
      ViewWlkTranU( view, -DefaultIncr );
      break;
     case K_d:       /* d or */
     case K_sRIGHT:  /* Shift-Right -> move observer right (in eye +U direc.) */
      ViewWlkTranU( view, DefaultIncr );
      break;
     case K_sDOWN:   /* Shift-Down -> move observer down (in -V direction) */
      ViewWlkTranV( view, -DefaultIncr );
      break;
     case K_sUP:     /* Shift-Up -> move observer up (in +V direction) */
      ViewWlkTranV( view, DefaultIncr );
      break;
     case K_cDOWN:   /* Crtl-Down -> turn head down (around eye -U axis) */
      ViewWlkRotU( view, -RotFactor );
      break;
     case K_cUP:     /* Crtl-Up -> turn head up (around eye +U axis) */
      ViewWlkRotU( view, RotFactor );
      break;
     default:
      return;
    }
   break;
  case 'o':  /* object-based model */
   if( view->control == 'r' )
   {
    switch( key )
    {
     case K_LEFT:   /* Left -> rotate object around -X axis about center */
      ViewObjRotX( view, -RotFactor );
      break;
     case K_RIGHT:  /* Right -> rotate object around +X axis about center */
      ViewObjRotX( view, RotFactor );
      break;
     case K_DOWN:   /* Down -> rotate object around -Y axis about center */
      ViewObjRotY( view, -RotFactor );
      break;
     case K_UP:     /* Up -> rotate object around +Y axis about center */
      ViewObjRotY( view, RotFactor );
      break;
     case K_sLEFT:  /* Shift-Left -> rotate obj. around -Z axis about center */
      ViewObjRotZ( view, -RotFactor );
      break;
     case K_sRIGHT: /* Shift-Right -> rotate obj. around +Z axis about center */
      ViewObjRotZ( view, RotFactor );
      break;
     default:
      return;
    }
   }
   else if( view->control == 't' )
   {
    switch( key )
    {
     case K_LEFT:   /* Left -> translate object in -X direction */
      ViewObjTranX( view, -DefaultIncr );
      break;
     case K_RIGHT:  /* Right -> translate object in +X direction */
      ViewObjTranX( view, DefaultIncr );
      break;
     case K_DOWN:   /* Down -> translate object in -Y direction */
      ViewObjTranY( view, -DefaultIncr );
      break;
     case K_UP:     /* Up -> translate object in +Y direction */
      ViewObjTranY( view, DefaultIncr );
      break;
     case K_sDOWN:  /* Shift-Left -> translate object in -Z direction */
      ViewObjTranZ( view, -DefaultIncr );
      break;
     case K_sUP: /* Shift-Right -> translate object in +Z direction */
      ViewObjTranZ( view, DefaultIncr );
      break;
     default:
      return;
    }
   }
   break;
  }
 }
}

/* ========================  v3dGetType  ====================== */
 
void v3dGetType( void *canvas, char *vtype )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 *vtype = view->type;
}

/* =======================  v3dGetModel  ====================== */
 
void v3dGetModel( void *canvas, char *vmodel )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 *vmodel = view->model;
}

/* ======================  v3dGetControl  ===================== */
 
void v3dGetControl( void *canvas, char *control )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 *control = view->control;
}

/* =====================  v3dGetViewVolume  ==================== */

void v3dGetViewVolume( void *canvas, double *left, double *right,
                                     double *bottom, double *top, 
                                     double *front, double *back )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 *left   = view->left;
 *right  = view->right;
 *bottom = view->bottom;
 *top    = view->top;
 *front  = view->front;
 *back   = view->back;
}

/* =======================  v3dGetCamera  ====================== */

void v3dGetCamera( void *canvas, double *eyex, double *eyey, double *eyez,
                                 double *refx, double *refy, double *refz,
                                 double *vupx, double *vupy, double *vupz )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 *eyex = view->eye.x;
 *eyey = view->eye.y;
 *eyez = view->eye.z;
 *refx = view->ref.x;
 *refy = view->ref.y;
 *refz = view->ref.z;
 *vupx = view->vup.x;
 *vupy = view->vup.y;
 *vupz = view->vup.z;
}

/* =====================  v3dSetViewVolume  ==================== */

void v3dSetViewVolume( void *canvas, double left, double right,
                                     double bottom, double top, 
                                     double front, double back )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

 view->left = left;
 view->right = right;
 view->bottom = bottom;
 view->top = top;
 view->front = front;
 view->back = back;

 AdjustWindow( view, &view->left, &view->right, &view->bottom, &view->top );
}

/* =======================  v3dSetCamera  ====================== */

void v3dSetCamera( void *canvas, double eyex, double eyey, double eyez,
                                 double refx, double refy, double refz,
                                 double vupx, double vupy, double vupz )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

 view->eye.x = eyex;
 view->eye.y = eyey;
 view->eye.z = eyez;
 view->ref.x = refx;
 view->ref.y = refy;
 view->ref.z = refz;
 view->vup.x = vupx;
 view->vup.y = vupy;
 view->vup.z = vupz;
}

/* ========================  v3dSeekRef  ======================= */

void v3dSeekRef( void *canvas, double refx, double refy, double refz )
{
 V3dState  *view;
 Vector    viewline;          /* current view direction          */
 double    d;                 /* distance from eye to ref. point */
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

/* Compute current view direction.
 */
 viewline.x = view->eye.x - view->ref.x;
 viewline.y = view->eye.y - view->ref.y;
 viewline.z = view->eye.z - view->ref.z;
 vecUnit( &viewline, &viewline );

/* Compute distance between eye and reference point.
 */
 d = Len3( view->eye.x - view->ref.x, 
           view->eye.y - view->ref.y, 
           view->eye.z - view->ref.z );

/* Update reference point position.
 */
 view->ref.x = refx;
 view->ref.y = refy;
 view->ref.z = refz;

/* Define eye position based on current view-line direction 
 * and distance to reference point.
 */
 view->eye.x = view->ref.x + d * viewline.x;
 view->eye.y = view->ref.y + d * viewline.y;
 view->eye.z = view->ref.z + d * viewline.z;
}

/* ==========================  v3dSave  ======================== */

void v3dSave( void *canvas )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->saved      = 1;
 view->type_sav   = view->type;
 view->eye_sav    = view->eye;
 view->ref_sav    = view->ref;
 view->vup_sav    = view->vup;
 view->left_sav   = view->left;
 view->right_sav  = view->right;
 view->bottom_sav = view->bottom;
 view->top_sav    = view->top;
 view->front_sav  = view->front;
 view->back_sav   = view->back;
}

/* ========================  v3dRestore  ======================= */

void v3dRestore( void *canvas )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

 view->type   = view->type_sav;
 view->eye    = view->eye_sav;
 view->ref    = view->ref_sav;
 view->vup    = view->vup_sav;
 view->left   = view->left_sav;
 view->right  = view->right_sav;
 view->bottom = view->bottom_sav;
 view->top    = view->top_sav;
 view->front  = view->front_sav;
 view->back   = view->back_sav;
 AdjustWindow( view, &view->left, &view->right, &view->bottom, &view->top );
}

/* ========================  v3dSaveFile  ====================== */

void v3dSaveFile( void *canvas, FILE *fd )
{
 V3dState  *view;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 if( view->type == 'o' )
  fprintf( fd, "%d\n", 0 );
 else
  fprintf( fd, "%d\n", 1 );
 fprintf( fd, "%g %g %g\n", view->eye.x, view->eye.y, view->eye.z );
 fprintf( fd, "%g %g %g\n", view->ref.x, view->ref.y, view->ref.z );
 fprintf( fd, "%g %g %g\n", view->vup.x, view->vup.y, view->vup.z );
 fprintf( fd, "%g %g\n", view->left, view->right );
 fprintf( fd, "%g %g\n", view->bottom, view->top );
 fprintf( fd, "%g %g\n", view->front, view->back );
}

/* ======================  v3dRestoreFile  ===================== */

void v3dRestoreFile( void *canvas, FILE *fd )
{
 V3dState  *view;
 int       type;
 
 view = GetViewState( canvas );
 if( view == NULL )
  return;

 view->update_type = V3dNoUpdate;

 if( fscanf( fd, "%d", &type ) != 1 )
  return;
 if( type == 0 )
  view->type = 'o';
 else
  view->type = 'p';
 fscanf( fd, "%lg %lg %lg", &view->eye.x, &view->eye.y, &view->eye.z );
 fscanf( fd, "%lg %lg %lg", &view->ref.x, &view->ref.y, &view->ref.z );
 fscanf( fd, "%lg %lg %lg", &view->vup.x, &view->vup.y, &view->vup.z );
 fscanf( fd, "%lg %lg", &view->left, &view->right );
 fscanf( fd, "%lg %lg", &view->bottom, &view->top );
 fscanf( fd, "%lg %lg", &view->front, &view->back );
 AdjustWindow( view, &view->left, &view->right, &view->bottom, &view->top );
}

