/*
** --------------------------------------------------------------
** v3d.h - 3D View Specification Package Header
**
** This header file contains the public function prototypes 
** for the 3D View Specification Package.
**
** ---------------------------------------------------------------
** Purpose:
** This package allows incremental updating of the view of a 
** model on a canvas window.
**
** ---------------------------------------------------------------
** Description:
**
** This module manipulates the view volume and viewing (camera)
** parameters through commands (function calls) and given input 
** mouse positions on a canvas window.  The parameters managed by 
** the module are:
**
**  Viewing parameters in client (object) coordinate system:
**   Point  eye             - eye (camera) position
**   Point  ref             - reference position
**   Point  vup             - up-vector (is a unit vector)
** 
**  Visualization volume (in eye coordinate system):
**   double front           - front/back clip planes:
**   double back              measured as dist. from eye
**   double left            - left view volume limit
**   double right           - right view volume limit
**   double bottom          - bottom view volume limit
**   double top             - top view volume limit
**
**  Generic parameters:
**   char   defvup          - default view up-vector:
**                            either 'x', 'y', or 'z'
**   char   type            - 'o' ==> orthographic
**                            'p' ==> perspective
**   char   model           - 's' ==> screen based
**                            'w' ==> walk through model
**                            'o' ==> object based
**   char   control         - 'r' ==> rotation control
**                            't' ==> translation control
**
** These parameters constitute a 3D view camera specification
** descriptor.  It is possible to have a number of these 
** descriptors.  To use this package, clients should first call
** function v3dOpen to create a descriptor.  This function 
** requires one argument, which is a client handle to the 
** target canvas.  This handle will be used as an identifier for 
** the viewing specification descriptor.  All calls of functions 
** of this module have the canvas handle as the first parameter 
** (argument). This will set the target viewing descriptor as the
** current active one.
**
** It is assumed that there is only one viewport for each canvas.
** Furthermore, it is assumed that the viewport always covers the
** entire canvas.  When there is a resize event for a canvas, 
** clients of this module should call function v3dResize, passing
** the target canvas sizes.
**
** ---------------------------------------------------------------
** Public functions:
** ---------------------------------------------------------------
**
** int    v3dOpen( void *canvas )
**
**   canvas    - client handle to target canvas               (in)
**
** This function initializes a viewing specification descriptor.
** It allocates memory for to the viewing descriptor and links 
** it with the given client canvas handle.
**
** ---------------------------------------------------------------
**
** void   v3dClose( void *canvas )
**
**   canvas    - client handle to target canvas               (in)
**
** This function frees the memory that had been allocated for 
** storing the internal viewing specification descriptor which
** is associated with the given canvas handle.
**
** ---------------------------------------------------------------
**
** void   v3dSetModelBox( void *canvas, 
**                        double xmin, double xmax,
**                        double ymin, double ymax,
**                        double zmin, double zmax )
**
**   canvas    - client handle to target canvas               (in)
**   xmin,xmax - model limiting x coordinates                 (in)
**   ymin,ymax - model limiting y coordinates                 (in)
**   zmin,zmax - model limiting z coordinates                 (in)
**
** This function registers with the viewing specification
** descriptor, which is associated with the given canvas handle, 
** the bounding box of the client model.
**
** ---------------------------------------------------------------
**
** void   v3dSetDefVup( void *canvas, char defvup )
**
**   canvas    - client handle to target canvas               (in)
**   defvup    - default view up-vector:
**                either 'x', 'y', or 'z'                     (in)
**
** This function specifies the default direction for the view
** up-vector.  If this function is not called, the package 
** initializes the view up-vector in the positive y direction.
**
** ---------------------------------------------------------------
**
** void   v3dSetType( void *canvas, char vtype )
**
**   canvas    - client handle to target canvas               (in)
**   vtype     - view type ('o' ==> orthographic, 
**                          'p' ==> perspective)              (in)
**
** This function changes the current view type according to the
** given type.  The package initializes the view type as 'p'
** (perspective).
**
** ---------------------------------------------------------------
**
** void   v3dSetModel( void *canvas, char vmodel )
**
**   canvas    - client handle to target canvas               (in)
**   vmodel    - 's' ==> screen based
**               'w' ==> walk through model
**               'o' ==> object based                         (in)
**
** This function changes the current viewing model according to 
** the given type.  The package initializes the model type as 's'
** (screen based).
**
** ---------------------------------------------------------------
**
** void   v3dSetControl( void *canvas, char control )
**
**   canvas    - client handle to target canvas               (in)
**   control   - 'r' ==> rotation control
**               't' ==> translation control                  (in)
**
** This function changes the current viewing control mechanism 
** according to the given type.  The package initializes the 
** control type as 'r' (rotation control).
**
** ---------------------------------------------------------------
**
** void   v3dDefault( void *canvas )
**
**   canvas    - client handle to target canvas               (in)
**
** This function sets up the default values for the visualization
** parameters according to the object/environment bounding box.
** The view direction is set according to the default view 
** up-vector direction.
**
** ---------------------------------------------------------------
**
** void   v3dResize( void *canvas, int width, int height )
**
**   canvas    - client handle to target canvas               (in)
**   width     - new horizontal canvas width in pixels        (in)
**   height    - new vertical canvas height in pixels         (in)
**
** This function should be called when the canvas whose handle is
** given is resized.  It takes the appropriate actions to reflect 
** the canvas resize.
**
** ---------------------------------------------------------------
**
** void   v3dZoomIn( void *canvas )
**
**   canvas    - client handle to target canvas               (in)
**
** This function adjusts the view window of the targed canvas 
** based on the current zoom factor.  It decreases the window 
** sizes, maitaining its current center.
** In this context, a window is the lateral view volume limits
** (left, right, bottom, top).
**
** ---------------------------------------------------------------
**
** void   v3dZoomOut( void *canvas )
**
**   canvas    - client handle to target canvas               (in)
**
** This function adjusts the view window of the targed canvas 
** based on the current zoom factor.  It increases the window 
** sizes, maitaining its current center.
** In this context, a window is the lateral view volume limits
** (left, right, bottom, top).
**
** ---------------------------------------------------------------
**
** void   v3dFit( void *canvas )
**
**   canvas    - client handle to target canvas               (in)
**
** This function sets up the values of the lateral view volume
** parameters (left, right, top, bottom) such that the entire
** model fits in the canvas viewport.
** The other visualization parameters (front, back, reference 
** point position, and eye position) are maintained.
**
** ---------------------------------------------------------------
**
** void   v3dProjPlane( void *canvas, char *plane )
**
**   canvas    - client handle to target canvas               (in)
**   plane     - string with the definition of project.plane  (in)
**
** This function changes the current values of the camera (eye)
** position vector and the view up-vector (vup) for the given 
** canvas to implement a projection of the client's model on the 
** target projection plane.
** Currently, the following projetion planes are available
** (view is vector from eye to ref.pt.):
**  plane = "xy", 
**  plane = "XY", 
**  plane = "+x+y", or
**  plane = "+X+Y"      --> view = -z     vup = +y;
**  plane = "xz", 
**  plane = "XZ", 
**  plane = "+x+z", or
**  plane = "+X+Z"      --> view = +y     vup = +z;
**  plane = "yz", 
**  plane = "YZ", 
**  plane = "+y+z", or
**  plane = "+Y+Z"      --> view = -x     vup = +z.
**  plane = "IsoXYZ",
**  plane = "Iso+X+Y+Z", or
**  plane = "Iso+x+y+z" --> view = (-1/sqrt(3),-1/sqrt(3),-1/sqrt(3)).
**                          vup  = (-1/srqt(6),-1/srqt(6),+2/sqrt(6)).
**  plane = "IsoZXY",
**  plane = "Iso+Z+X+Y", or
**  plane = "Iso+z+x+y" --> view = (-1/sqrt(3),-1/sqrt(3),-1/sqrt(3)).
**                          vup  = (-1/srqt(6),+2/srqt(6),-1/sqrt(6)).
**  plane = "IsoYZX",
**  plane = "Iso+Y+Z+X", or
**  plane = "Iso+y+z+x" --> view = (-1/sqrt(3),-1/sqrt(3),-1/sqrt(3)).
**                          vup  = (+2/srqt(6),-1/srqt(6),-1/sqrt(6)).
**
** ---------------------------------------------------------------
**
** void   v3dMouseBt( void *canvas, int b, int mode, int x, int y, 
**                    char *mousecfg )
**
**   canvas   - client handle to target canvas                (in)
**   b        - button of mouse that is pressed               (in)
**   x, y     - mouse position in device (canvas) coordinates (in)
**   mousecfg - keyboard configuration when mouse is pressed  (in)
**
** This is a callback function for a mouse button event at at 
** canvas.
** At button pressed mode (mode = 0),this function sets ups the 
** viewing package for the next mouse movement.
** At button released mode (mode = 1), this function finishes up
** visualization control by the mouse.
**
** In any control mechanism, if the Control key is pressed, 
** left mouse button controls the viewing scaling factor (control
** of size of visualization window) and right mouse button
** controls the distance between the eye point and the projection
** plane.  If the Shift key is pressed, left mouse button controls
** the pan (translation of the visualization window).
**
** Otherwise, 
** If the current control mechanism is rotation control (see 
** function v3dSetControl), the viewing is changed in the 
** following way:
** * If the current viewing model is screen-based or walk-through 
**   (see function v3dSetModel), left mouse button controls x and 
**   y screen axis rotation and right mouse button controls z 
**   screen axis rotation.
**
** Else if the current control mechanism is translation control 
** (see function v3dSetControl), the viewing is changed in the 
** following way:
** * If the current viewing model is screen-based or walk-through 
**   (see function v3dSetModel), left mouse button controls x and 
**   y screen axis translation and right mouse button controls z 
**   screen axis translation.
**
** ---------------------------------------------------------------
**
** void   v3dMouseMv( void *canvas, int x, int y, char *mousecfg )
**
**   canvas   - client handle to target canvas                (in)
**   x, y     - mouse position in device (canvas) coordinates (in)
**   mousecfg - keyboard configuration when mouse is pressed  (in)
**
** This function changes the viewing parameters according to the
** given mouse movement, that is, according to given mouse 
** position with respect to its position specified in the last 
** call to function ViewSetup.
**
** ---------------------------------------------------------------
**
** void   v3dIncrCrtl( void *canvas, char axis, char dir )
**
**   canvas    - client handle to target canvas               (in)
**   axis      - defines axis (either 'x', 'y', or 'z')       (in)
**   dir       - defines direction (either '-' or '+')        (in)
**
** This function performes incremental changes the viewing 
** parameters according the given axis and direction.
** The increments will apply to either rotation or translation
** depending on the current control mechanism (see function 
** v3dSetControl).
** The changes will depend on the current visualization model:
** screen based, walk-through, or object based (see function 
** v3dSetModel). 
**
** ---------------------------------------------------------------
**
** void   v3dKeyCrtl( void *canvas, int key )
**
**   canvas    - client handle to target canvas               (in)
**   key       - keyboard key touched                         (in)
**
** This function changes the viewing parameters according to a 
** keyboard input.
**
** ---------------------------------------------------------------
**
** void   v3dGetType( void *canvas, char *vtype )
**
**   canvas    - client handle to target canvas               (in)
**   vtype     - view type ('o' ==> orthographic, 
**                          'p' ==> perspective)             (out)
**
** This function returns the view type of the given canvas viewing 
** specification descriptor.
**
** ---------------------------------------------------------------
**
** void   v3dGetModel( void *canvas, char *vmodel )
**
**   canvas    - client handle to target canvas               (in)
**   vmodel    - 's' ==> screen based
**               'w' ==> walk through model
**               'o' ==> object based                        (out)
**
** This function returns the current viewing model associated to
** the given canvas.
**
** ---------------------------------------------------------------
**
** void   v3dGetControl( void *canvas, char *control )
**
**   canvas    - client handle to target canvas               (in)
**   control   - 'r' ==> rotation control
**               't' ==> translation control                 (out)
**
** This function returns the current viewing control mechanism 
** associated to the given canvas.
**
** ---------------------------------------------------------------
**
** void   v3dGetViewVolume( void *canvas, double *left, double *right,
**                                        double *bottom, double *top, 
**                                        double *front, double *back )
**
**   canvas    - client handle to target canvas               (in)
**   left, right, bottom, top, front, back - 
**            - current visualization volume                 (out)
**
** This funtion returns the visualization volume of the given canvas
** viewing specification descriptor.
**
** ---------------------------------------------------------------
**
** void   v3dGetCamera( void *canvas,
**                      double *eyex, double *eyey, double *eyez,
**                      double *refx, double *refy, double *refz,
**                      double *vupx, double *vupy, double *vupz )
**
**   canvas    - client handle to target canvas               (in)
**   eyex,eyey,eyez - camera (eye) position in client space  (out)
**   refx,refy,refz - reference pt.position in client space  (out)
**   vupx,vupy,vupz - view up vector coordinates             (out)
**
** This funtion returns the camera visualization parameters of the 
** given canvas viewing specification descriptor.
**
** ---------------------------------------------------------------
**
** void   v3dSetViewVolume( void *canvas, double left, double right,
**                                        double bottom, double top, 
**                                        double front, double back )
**
**   canvas    - client handle to target canvas               (in)
**   left, right, bottom, top, front, back - 
**            - current visualization volume                  (in)
**
** This funtion sets the visualization volume of the given canvas
** viewing specification descriptor.
**
** ---------------------------------------------------------------
**
** void   v3dSetCamera( void *canvas,
**                      double eyex, double eyey, double eyez,
**                      double refx, double refy, double refz,
**                      double vupx, double vupy, double vupz )
**
**   canvas    - client handle to target canvas               (in)
**   eyex,eyey,eyez - camera (eye) position in client space   (in)
**   refx,refy,refz - reference pt.position in client space   (in)
**   vupx,vupy,vupz - view up vector coordinates              (in)
**
** This funtion sets the camera visualization parameters of the 
** given canvas viewing specification descriptor.
**
** ---------------------------------------------------------------
**
** void   v3dSeekRef( void *canvas, 
**                    double refx, double refy, double refz )
**
**   canvas    - client handle to target canvas               (in)
**   refx,refy,refz - reference pt.position in client space   (in)
**
** This funtion sets the camera visualization reference point of the 
** given canvas viewing specification descriptor.  The current 
** view-line direction (from eye to reference point), the 
** distance between the eye point to reference point, and the 
** view up-vector are maintained.  The view volume parameters 
** (left, right, bottom, top, front, back) are also maintained.
**
** ---------------------------------------------------------------
**
** void   v3dSave( void *canvas )
**
**   canvas    - client handle to target canvas               (in)
**
** This funtion saves all the viewing parameters associated
** to the given canvas in its viewing descriptor.
**
** ---------------------------------------------------------------
**
** void   v3dRestore( void *canvas )
**
**   canvas    - client handle to target canvas               (in)
**
** This funtion restores all the viewing parameters associated
** to the given canvas from the parameters saved in its descriptor.
**
** ---------------------------------------------------------------
**
** void   v3dSaveFile( void *canvas, FILE *fd )
**
**   canvas    - client handle to target canvas               (in)
**   fd        - file descriptor                              (in)
**
** This funtion saves all the viewing parameters associated
** to the given canvas in the file given by its descriptor.
**
** ---------------------------------------------------------------
**
** void   v3dRestoreFile( void *canvas, FILE *fd )
**
**   canvas    - client handle to target canvas               (in)
**   fd        - file descriptor                              (in)
**
** This funtion restores all the viewing parameters associated
** to the given canvas from the file given by its descriptor.
**
** ---------------------------------------------------------------
**
** Version:    1.0
**
** Created:    08-Nov-2006    Luiz F. Martha
**    Stolen from V3d module (Tecgraf/PUC-Rio).
**
*/

#ifndef _V3D_H
#define _V3D_H

#include <stdio.h>  /* FILE */

#ifdef __cplusplus
extern "C" {
#endif


int  v3dOpen       ( void *canvas );
void v3dClose      ( void *canvas );
void v3dSetModelBox( void *canvas, 
                     double xmin, double xmax,
                     double ymin, double ymax,
                     double zmin, double zmax );
void v3dSetDefVup  ( void *canvas, char defvup );
void v3dSetType    ( void *canvas, char vtype );
void v3dSetModel   ( void *canvas, char vmodel );
void v3dSetControl ( void *canvas, char control );
void v3dDefault    ( void *canvas );
void v3dResize     ( void *canvas, int width, int height );
void v3dZoomIn     ( void *canvas );
void v3dZoomOut    ( void *canvas );
void v3dFit        ( void *canvas );
void v3dProjPlane  ( void *canvas, char *plane );
void v3dMouseBt    ( void *canvas, int b, int mode, int x, int y, 
                                   char *mousecfg );
void v3dMouseMv    ( void *canvas, int x, int y, char *mousecfg );
void v3dIncrCrtl   ( void *canvas, char axis, char dir );
void v3dKeyCrtl    ( void *canvas, int key );
void v3dGetType    ( void *canvas, char *vtype );
void v3dGetModel   ( void *canvas, char *vmodel );
void v3dGetControl ( void *canvas, char *control );
void v3dGetViewVolume( void *canvas, 
                                 double *left, double *right, double *bottom, 
                                 double *top, double *front, double *back );
void v3dGetCamera    ( void *canvas, 
                                 double *eyex, double *eyey, double *eyez, 
                                 double *refx, double *refy, double *refz,
                                 double *vupx, double *vupy, double *vupz );
void v3dSetViewVolume( void *canvas, 
                                 double left, double right, double bottom, 
                                 double top, double front, double back );
void v3dSetCamera    ( void *canvas, 
                                 double eyex, double eyey, double eyez, 
                                 double refx, double refy, double refz,
                                 double vupx, double vupy, double vupz );
void v3dSeekRef    ( void *canvas, double refx, double refy, double refz );
void v3dSave       ( void *canvas );
void v3dRestore    ( void *canvas );
void v3dSaveFile   ( void *canvas, FILE *fd );
void v3dRestoreFile( void *canvas, FILE *fd );

#ifdef __cplusplus
}
#endif

#endif
